import logging
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np
import torch
from torch import nn

import flyvis
from flyvis.utils.logging_utils import warn_once

logger = logging.getLogger(__name__)


def recover_network(
    network: nn.Module,
    state_dict: Union[Dict, Path, str],
    ensemble_and_network_id: str = None,
) -> nn.Module:
    """
    Load network parameters from state dict.

    Args:
        network: FlyVision network.
        state_dict: State or path to checkpoint containing the "network" parameters.
        ensemble_and_network_id: Optional identifier for the network.

    Returns:
        The updated network.
    """
    state = get_from_state_dict(state_dict, "network")
    if state is not None:
        network.load_state_dict(state)
        logging.info(
            "Recovered network state%s",
            f" {ensemble_and_network_id}." if ensemble_and_network_id else ".",
        )
    else:
        logging.warning("Could not recover network state.")
    return network


def recover_decoder(
    decoder: Dict[str, nn.Module], state_dict: Union[Dict, Path], strict: bool = True
) -> Dict[str, nn.Module]:
    """
    Recover multiple decoders from state dict.

    Args:
        decoder: Dictionary of decoders.
        state_dict: State or path to checkpoint.
        strict: Whether to strictly enforce that the keys in state_dict match.

    Returns:
        The updated dictionary of decoders.
    """
    states = get_from_state_dict(state_dict, "decoder")
    if states is not None:
        for key, dec in decoder.items():
            state = states.pop(key, None)
            if state is not None:
                dec.load_state_dict(state, strict=strict)
                logging.info("Recovered %s decoder state.", key)
            else:
                logging.warning("Could not recover state of %s decoder.", key)
    else:
        logging.warning("Could not recover decoder states.")
    return decoder


def recover_optimizer(
    optimizer: torch.optim.Optimizer, state_dict: Union[Dict, Path]
) -> torch.optim.Optimizer:
    """
    Recover optimizer state from state dict.

    Args:
        optimizer: PyTorch optimizer.
        state_dict: State or path to checkpoint.

    Returns:
        The updated optimizer.
    """
    state = get_from_state_dict(state_dict, "optim")
    if state is not None:
        optimizer.load_state_dict(state)
        logging.info("Recovered optimizer state.")
    else:
        logging.warning("Could not recover optimizer state.")
    return optimizer


def recover_penalty_optimizers(
    optimizers: Dict[str, torch.optim.Optimizer], state_dict: Union[Dict, Path]
) -> Dict[str, torch.optim.Optimizer]:
    """
    Recover penalty optimizers from state dict.

    Args:
        optimizers: Dictionary of penalty optimizers.
        state_dict: State or path to checkpoint.

    Returns:
        The updated dictionary of penalty optimizers.
    """
    states = get_from_state_dict(state_dict, "penalty_optims")
    if states is not None:
        for key, optim in optimizers.items():
            state = states.pop(key, None)
            if state is not None:
                optim.load_state_dict(state)
                logging.info("Recovered %s optimizer state.", key)
            else:
                logging.warning("Could not recover state of %s optimizer.", key)
    else:
        logging.warning("Could not recover penalty optimizer states.")
    return optimizers


def get_from_state_dict(state_dict: Union[Dict, Path, str], key: str) -> Dict:
    """
    Get a specific key from the state dict.

    Args:
        state_dict: State dict or path to checkpoint.
        key: Key to retrieve from the state dict.

    Returns:
        The value associated with the key in the state dict.

    Raises:
        TypeError: If state_dict is not of type Path, str, or dict.
    """
    if state_dict is None:
        return None
    if isinstance(state_dict, (Path, str)):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=FutureWarning)
            state = torch.load(
                state_dict, map_location=flyvis.device, weights_only=False
            ).pop(key, None)
    elif isinstance(state_dict, dict):
        state = state_dict.get(key, None)
    else:
        raise TypeError(
            f"state_dict must be of type Path, str or dict, but is {type(state_dict)}."
        )
    return state


@dataclass
class Checkpoints:
    """
    Dataclass to store checkpoint information.

    Attributes:
        indices: List of checkpoint indices.
        paths: List of checkpoint paths.
    """

    indices: List[int]
    paths: List[Path]

    def __repr__(self):
        return (
            f"Checkpoints(\n"
            f"  indices={repr(self.indices)},\n"
            f"  paths={repr(self.paths)},\n"
            f")"
        )


def resolve_checkpoints(
    networkdir: "flyvis.network.NetworkDir",
) -> Checkpoints:
    """
    Resolve checkpoints from network directory.

    Args:
        networkdir: FlyVision network directory.

    Returns:
        A Checkpoints object containing indices and paths of checkpoints.
    """
    indices, paths = checkpoint_index_to_path_map(networkdir.chkpts.path)
    return Checkpoints(indices, paths)


def checkpoint_index_to_path_map(
    path: Path, glob: str = "chkpt_*"
) -> Tuple[List[int], List[Path]]:
    """
    Returns all numerical identifiers and paths to checkpoints stored in path.

    Args:
        path: Checkpoint directory.
        glob: Glob pattern for checkpoint files.

    Returns:
        A tuple containing a list of indices and a list of paths to checkpoints.
    """
    import re

    path.mkdir(exist_ok=True)
    paths = np.array(sorted(list((path).glob(glob))))
    try:
        _index = [int(re.findall(r"\d{1,10}", p.parts[-1])[0]) for p in paths]
        _sorting_index = np.argsort(_index)
        paths = paths[_sorting_index].tolist()
        index = np.array(_index)[_sorting_index].tolist()
        return index, paths
    except IndexError:
        return [], paths


def best_checkpoint_default_fn(
    path: Path,
    validation_subdir: str = "validation",
    loss_file_name: str = "loss",
) -> Path:
    """
    Find the best checkpoint based on the minimum loss.

    Args:
        path: Path to the network directory.
        validation_subdir: Subdirectory containing validation data.
        loss_file_name: Name of the loss file.

    Returns:
        Path to the best checkpoint.
    """
    networkdir = flyvis.NetworkDir(path)
    checkpoint_dir = networkdir.chkpts.path
    indices, paths = checkpoint_index_to_path_map(checkpoint_dir, glob="chkpt_*")
    loss_file_name = check_loss_name(networkdir[validation_subdir], loss_file_name)
    index = np.argmin(networkdir[validation_subdir][loss_file_name][()])
    index = indices[index]
    path = paths[index]
    return path


def check_loss_name(loss_folder, loss_file_name: str) -> str:
    """
    Check if the loss file name exists in the loss folder.

    Args:
        loss_folder: The folder containing loss files.
        loss_file_name: The name of the loss file to check.

    Returns:
        The validated loss file name.
    """
    if loss_file_name not in loss_folder and "loss" in loss_folder:
        warn_once(
            logging,
            f"{loss_file_name} not in {loss_folder.path}, but 'loss' is. "
            "Falling back to 'loss'. You can rerun the ensemble validation to make "
            "appropriate recordings of the losses.",
        )
        loss_file_name = "loss"
    return loss_file_name


if __name__ == "__main__":
    nv = flyvis.NetworkView("flow/9998/000")
    print(resolve_checkpoints(nv.dir))



"""Base classes for all dynamic stimuli datasets."""

from contextlib import contextmanager
from typing import Any, Dict, Iterable, List, Union

import numpy as np
import pandas as pd
import torch

from flyvis.datasets.augmentation.temporal import get_temporal_sample_indices
from flyvis.utils.dataset_utils import get_random_data_split
from flyvis.utils.df_utils import where_dataframe

__all__ = ["SequenceDataset", "StimulusDataset", "MultiTaskDataset"]


class SequenceDataset(torch.utils.data.Dataset):
    """Base class for all sequence datasets.

    All sequence datasets can subclass this class. They are expected to implement
    the following attributes and methods.

    Attributes:
        framerate (int): Framerate of the original sequences.
        dt (float): Sampling and integration time constant.
        t_pre (float): Warmup time.
        t_post (float): Cooldown time.
        arg_df (pd.DataFrame): required DataFrame containing the dataset parameters.
    """

    arg_df: pd.DataFrame = None
    dt: float = None
    t_pre: float = None
    t_post: float = None

    def get_item(self, key: int) -> Any:
        """Return an item of the dataset.

        Args:
            key: Index of the item to retrieve.

        Returns:
            The dataset item at the specified index.
        """

    def __len__(self) -> int:
        """Size of the dataset."""
        return len(self.arg_df)

    def __getitem__(self, key: Union[slice, Iterable, int, np.int_]) -> Any:
        """Implements advanced indexing.

        Args:
            key: Index, slice, or iterable of indices.

        Returns:
            The dataset item(s) at the specified index/indices.

        Raises:
            IndexError: If the index is out of range.
            TypeError: If the key type is invalid.
        """
        return getitem(self, key)

    def get_temporal_sample_indices(
        self, n_frames: int, total_seq_length: int, augment: bool = None
    ) -> torch.Tensor:
        """Returns temporal indices to sample from a sequence.

        Args:
            n_frames: Number of sequence frames to sample from.
            total_seq_length: Total sequence length.
            augment: If True, picks the start frame at random. If False, starts at 0.

        Returns:
            Tensor of temporal indices.

        Note:
            Interpolates between start_index and start_index + n_frames and rounds the
            resulting float values to integer to create indices. This can lead to
            irregularities in terms of how many times each raw data frame is sampled.
        """
        augment = augment if augment is not None else getattr(self, "augment", False)
        framerate = getattr(self, "original_framerate", 1 / self.dt)
        return get_temporal_sample_indices(
            n_frames, total_seq_length, framerate, self.dt, augment
        )


class StimulusDataset(SequenceDataset):
    """Base class for stimulus datasets."""

    def get_stimulus_index(self, kwargs: Dict[str, Any]) -> int:
        """Get the sequence id for a set of arguments.

        Args:
            kwargs: Dictionary containing independent arguments or parameters
                describing the sample of the dataset.

        Returns:
            The sequence id for the given arguments.

        Raises:
            ValueError: If arg_df attribute is not specified.

        Note:
            The child dataset implements the specific method:
            ```python
            def get_stimulus_index(self, arg1, arg2, ...):
                return StimulusDataset.get_stimulus_index(locals())
            ```
            with locals() specifying kwargs in terms of `arg1`, `arg2`, ...
            to index arg_df.
        """
        if getattr(self, "arg_df", None) is None:
            raise ValueError("arg_df attribute not specified.")

        if "self" in kwargs:
            del kwargs["self"]

        return where_dataframe(self.arg_df, **kwargs).item()


class MultiTaskDataset(SequenceDataset):
    """Base class for all (multi-)task sequence datasets.

    All (multi-)task sequence datasets can subclass this class. They are expected
    to implement the following additional attributes and methods.

    Attributes:
        tasks (List[str]): A list of all tasks.
        augment (bool): Turns augmentation on and off.
    """

    tasks: List[str] = []
    augment: bool = False

    @contextmanager
    def augmentation(self, abool: bool) -> None:
        """Contextmanager to turn augmentation on or off in a code block.

        Args:
            abool: Boolean value to set augmentation.

        Example:
            ```python
            with dataset.augmentation(True):
                for i, data in enumerate(dataloader):
                    ...  # all data is augmented
            ```
        """
        _prev = self.augment
        self.augment = abool
        try:
            yield
        finally:
            self.augment = _prev

    def get_random_data_split(
        self, fold: int, n_folds: int, shuffle: bool = True, seed: int = 0
    ) -> np.ndarray:
        """Returns a random data split.

        Args:
            fold: Current fold number.
            n_folds: Total number of folds.
            shuffle: Whether to shuffle the data.
            seed: Random seed for reproducibility.

        Returns:
            Array of indices for the data split.
        """
        return get_random_data_split(
            fold,
            n_samples=len(self),
            n_folds=n_folds,
            shuffle=shuffle,
            seed=seed,
        )


def getitem(cls: Any, key: Union[slice, Iterable, int, np.int_]) -> Any:
    """Implements advanced indexing for a dataset.

    Args:
        cls: The dataset instance.
        key: Index, slice, or iterable of indices.

    Returns:
        The item(s) at the specified index/indices.

    Raises:
        IndexError: If the index is out of range.
        TypeError: If the key type is invalid.
    """
    if isinstance(key, slice):
        return [cls[ii] for ii in range(*key.indices(len(cls)))]
    elif isinstance(key, Iterable):
        return [cls[i] for i in key]
    elif isinstance(key, (int, np.int_)):
        if key < 0:
            key += len(cls)
        if key < 0 or key >= len(cls):
            raise IndexError(f"The index ({key}) is out of range.")
        return cls.get_item(key)
    else:
        raise TypeError("Invalid argument type.")



import numpy as np
import pytest
import torch

from flyvis.datasets import rendering


@pytest.fixture(scope="module")
def boxeye() -> rendering.BoxEye:
    return rendering.BoxEye(15, 13)


def test_init(boxeye: rendering.BoxEye):
    assert boxeye.extent == 15
    assert boxeye.kernel_size == 13
    assert boxeye.receptor_centers.shape == (boxeye.hexals, 2)
    assert boxeye.min_frame_size.shape == (2,)
    assert boxeye.pad == (6, 6, 6, 6)


def test_call(boxeye: rendering.BoxEye):
    sequence = torch.ones((2, 2, 100, 100))

    rendered = boxeye(sequence, ftype="mean", hex_sample=True)
    assert rendered.shape == (2, 2, 1, boxeye.hexals)
    assert np.isclose(rendered.cpu().numpy().mean(), 1, atol=0.05)

    rendered = boxeye(sequence, ftype="sum", hex_sample=True)
    assert rendered.shape == (2, 2, 1, boxeye.hexals)
    assert np.isclose(rendered.cpu().numpy().mean(), boxeye.kernel_size**2, atol=5)

    sequence = torch.ones((2, 2, 100, 100)).random_(0, 11)

    rendered = boxeye(sequence, ftype="median", hex_sample=True)
    assert rendered.shape == (2, 2, 1, boxeye.hexals)
    assert np.isclose(rendered.cpu().numpy().mean(), 5, atol=0.05)

    rendered = boxeye(sequence.clone(), ftype="median", hex_sample=False)
    assert rendered.shape == (*sequence.shape[:2], *boxeye.min_frame_size.cpu().numpy())
    assert np.isclose(rendered.cpu().numpy().mean(), 5, atol=0.05)

    with pytest.raises(ValueError):
        boxeye(sequence, ftype="invalid", hex_sample=True)


def test_hex_render(boxeye: rendering.BoxEye):
    sequence = torch.ones((2, 2, 100, 100))
    rendered = boxeye.hex_render(sequence)
    assert rendered.shape == (2, 2, 1, boxeye.hexals)



from itertools import product
from typing import List, Literal, Optional

import numpy as np
import pandas as pd
import torch
from datamate import Namespace

import flyvis
from flyvis.utils import hex_utils
from flyvis.utils.hex_utils import Hexal, HexLattice

from .datasets import StimulusDataset
from .rendering.utils import pad, resample

__all__ = ["Dots", "CentralImpulses", "SpatialImpulses"]


class Dots(StimulusDataset):
    """
    Render flashes aka dots per ommatidia.

    Note:
        Renders directly in receptor space, does not use BoxEye or HexEye as eye-model.

    Args:
        dot_column_radius: Radius of the dot column.
        max_extent: Maximum extent of the stimulus.
        bg_intensity: Background intensity.
        t_stim: Stimulus duration.
        dt: Time step.
        t_impulse: Impulse duration.
        n_ommatidia: Number of ommatidia.
        t_pre: Pre-stimulus duration.
        t_post: Post-stimulus duration.
        intensity: Stimulus intensity.
        mode: Stimulus mode ('sustained' or 'impulse').
        device: Torch device for computations.

    Attributes:
        dt: Time step.
        arg_df: DataFrame containing stimulus parameters.
        config: Namespace containing configuration parameters.
        t_stim: Stimulus duration.
        n_ommatidia: Number of ommatidia.
        offsets: Array of ommatidia offsets.
        u: U-coordinates of the hexagonal grid.
        v: V-coordinates of the hexagonal grid.
        extent_condition: Boolean mask for the extent condition.
        max_extent: Maximum extent of the stimulus.
        bg_intensity: Background intensity.
        intensities: List of stimulus intensities.
        device: Torch device for computations.
        mode: Stimulus mode ('sustained' or 'impulse').
        params: List of stimulus parameters.
        t_impulse: Impulse duration.

    Raises:
        ValueError: If dot_column_radius is greater than max_extent.
    """

    dt: Optional[float] = None
    arg_df: Optional[pd.DataFrame] = None

    def __init__(
        self,
        dot_column_radius: int = 0,
        max_extent: int = 15,
        bg_intensity: float = 0.5,
        t_stim: float = 5,
        dt: float = 1 / 200,
        t_impulse: Optional[float] = None,
        n_ommatidia: int = 721,
        t_pre: float = 2.0,
        t_post: float = 0,
        intensity: float = 1,
        mode: Literal["sustained", "impulse"] = "sustained",
        device: torch.device = flyvis.device,
    ):
        if dot_column_radius > max_extent:
            raise ValueError("dot_column_radius must be smaller than max_extent")
        self.config = Namespace(
            dot_column_radius=dot_column_radius,
            max_extent=max_extent,
            bg_intensity=bg_intensity,
            t_stim=t_stim,
            dt=dt,
            n_ommatidia=n_ommatidia,
            t_pre=t_pre,
            t_post=t_post,
            intensity=intensity,
            mode=mode,
            t_impulse=t_impulse,
        )

        self.t_stim = t_stim
        self.t_pre = t_pre
        self.t_post = t_post

        self.n_ommatidia = n_ommatidia
        self.offsets = np.arange(self.n_ommatidia)

        u, v = hex_utils.get_hex_coords(hex_utils.get_hextent(n_ommatidia))
        extent_condition = (
            (-max_extent <= u)
            & (u <= max_extent)
            & (-max_extent <= v)
            & (v <= max_extent)
            & (-max_extent <= u + v)
            & (u + v <= max_extent)
        )
        self.u = u[extent_condition]
        self.v = v[extent_condition]
        # self.offsets = self.offsets[extent_condition]
        self.extent_condition = extent_condition

        # to have multi column dots at every location, construct coordinate_indices
        # for each central column
        coordinate_indices = []
        for u, v in zip(self.u, self.v):
            ring = HexLattice.filled_circle(
                radius=dot_column_radius, center=Hexal(u, v, 0), as_lattice=True
            )
            # mask = np.array([~np.isnan(h.value) for h in h1])
            coordinate_indices.append(self.offsets[ring.where(1)])

        self.max_extent = max_extent
        self.bg_intensity = bg_intensity

        self.intensities = [2 * bg_intensity - intensity, intensity]
        self.device = device
        self.mode = mode

        self.params = [
            (*p[0], p[-1])
            for p in list(
                product(
                    zip(self.u, self.v, self.offsets, coordinate_indices),
                    self.intensities,
                )
            )
        ]

        self.arg_df = pd.DataFrame(
            self.params,
            columns=["u", "v", "offset", "coordinate_index", "intensity"],
        )

        self.dt = dt
        self.t_impulse = t_impulse or self.dt

    def _params(self, key: int) -> np.ndarray:
        """Get parameters for a specific key.

        Args:
            key: Index of the parameters to retrieve.

        Returns:
            Array of parameters for the given key.
        """
        return self.arg_df.iloc[key].values

    def get_item(self, key: int) -> torch.Tensor:
        """Get a stimulus item for a specific key.

        Args:
            key: Index of the item to retrieve.

        Returns:
            Tensor representing the stimulus sequence.
        """
        # create maps with background value
        _dot = (
            torch.ones(self.n_ommatidia, device=self.device)[None, None]
            * self.bg_intensity
        )
        # fill at the ommatitdium at offset with intensity
        _, _, _, coordinate_index, intensity = self._params(key)
        _dot[:, :, coordinate_index] = torch.tensor(intensity, device=self.device).float()

        # repeat for stustained stimulus
        if self.mode == "sustained":
            sequence = resample(_dot, self.t_stim, self.dt, dim=1, device=self.device)

        elif self.mode == "impulse":
            # pad remaining stimulus duration i.e. self.t_stim - self.dt with
            # background intensity
            if self.t_impulse == self.dt:
                sequence = pad(
                    _dot,
                    self.t_stim,
                    self.dt,
                    mode="end",
                    fill=self.bg_intensity,
                )
            # first resample for t_impulse/dt then pad remaining stimulus
            # duration, i.e. t_stim - t_impulse with background intensity
            else:
                sequence = resample(
                    _dot, self.t_impulse, self.dt, dim=1, device=self.device
                )
                sequence = pad(
                    sequence,
                    self.t_stim,
                    self.dt,
                    mode="end",
                    fill=self.bg_intensity,
                )

        # pad with pre stimulus background
        sequence = pad(
            sequence,
            self.t_stim + self.t_pre,
            self.dt,
            mode="start",
            fill=self.bg_intensity,
        )
        # pad with post stimulus background
        sequence = pad(
            sequence,
            self.t_stim + self.t_pre + self.t_post,
            self.dt,
            mode="end",
            fill=self.bg_intensity,
        )
        return sequence.squeeze()

    def get_stimulus_index(self, u: float, v: float, intensity: float) -> int:
        """Get sequence ID from given arguments.

        Args:
            u: U-coordinate.
            v: V-coordinate.
            intensity: Stimulus intensity.

        Returns:
            Sequence ID.
        """
        return StimulusDataset.get_stimulus_index(self, locals())


class CentralImpulses(StimulusDataset):
    """Flashes at the center of the visual field for temporal receptive field mapping.

    Args:
        impulse_durations: List of impulse durations.
        dot_column_radius: Radius of the dot column.
        bg_intensity: Background intensity.
        t_stim: Stimulus duration.
        dt: Time step.
        n_ommatidia: Number of ommatidia.
        t_pre: Pre-stimulus duration.
        t_post: Post-stimulus duration.
        intensity: Stimulus intensity.
        mode: Stimulus mode.
        device: Torch device for computations.

    Attributes:
        arg_df: DataFrame containing stimulus parameters.
        dt: Time step.
        dots: Instance of the Dots class.
        impulse_durations: List of impulse durations.
        config: Configuration namespace.
        params: List of stimulus parameters.
    """

    arg_df: Optional[pd.DataFrame] = None
    dt: Optional[float] = None

    def __init__(
        self,
        impulse_durations: List[float] = [5e-3, 20e-3, 50e-3, 100e-3, 200e-3, 300e-3],
        dot_column_radius: int = 0,
        bg_intensity: float = 0.5,
        t_stim: float = 5,
        dt: float = 0.005,
        n_ommatidia: int = 721,
        t_pre: float = 2.0,
        t_post: float = 0,
        intensity: float = 1,
        mode: str = "impulse",
        device: torch.device = flyvis.device,
    ):
        """Initialize the CentralImpulses dataset.

        Args:
            impulse_durations: List of impulse durations.
            dot_column_radius: Radius of the dot column.
            bg_intensity: Background intensity.
            t_stim: Stimulus duration.
            dt: Time step.
            n_ommatidia: Number of ommatidia.
            t_pre: Pre-stimulus duration.
            t_post: Post-stimulus duration.
            intensity: Stimulus intensity.
            mode: Stimulus mode.
            device: Torch device for computations.
        """
        self.dots = Dots(
            dot_column_radius=dot_column_radius,
            max_extent=dot_column_radius,
            bg_intensity=bg_intensity,
            t_stim=t_stim,
            dt=dt,
            n_ommatidia=n_ommatidia,
            t_pre=t_pre,
            t_post=t_post,
            intensity=intensity,
            mode=mode,
            device=device,
        )
        self.impulse_durations = impulse_durations
        self.config = self.dots.config
        self.config.update(impulse_durations=impulse_durations)
        self.params = [
            (*p[0], p[1])
            for p in product(self.dots.arg_df.values.tolist(), impulse_durations)
        ]
        self.arg_df = pd.DataFrame(
            self.params,
            columns=[
                "u",
                "v",
                "offset",
                "coordinate_index",
                "intensity",
                "t_impulse",
            ],
        )
        self.dt = dt

    def _params(self, key: int) -> np.ndarray:
        """Get parameters for a specific key.

        Args:
            key: Index of the parameters to retrieve.

        Returns:
            Array of parameters for the given key.
        """
        return self.arg_df.iloc[key].values

    def get_item(self, key: int) -> torch.Tensor:
        """Get a stimulus item for a specific key.

        Args:
            key: Index of the item to retrieve.

        Returns:
            Tensor representing the stimulus sequence.
        """
        u, v, offset, coordinate_index, intensity, t_impulse = self._params(key)
        self.dots.t_impulse = t_impulse
        return self.dots[self.dots.get_stimulus_index(u, v, intensity)]

    @property
    def t_pre(self) -> float:
        """Get pre-stimulus duration."""
        return self.dots.t_pre

    @property
    def t_post(self) -> float:
        """Get post-stimulus duration."""
        return self.dots.t_post

    def __repr__(self) -> str:
        """Get string representation of the dataset."""
        return repr(self.arg_df)


class SpatialImpulses(StimulusDataset):
    """Spatial flashes for spatial receptive field mapping.

    Args:
        impulse_durations: List of impulse durations.
        max_extent: Maximum extent of the stimulus.
        dot_column_radius: Radius of the dot column.
        bg_intensity: Background intensity.
        t_stim: Stimulus duration.
        dt: Time step.
        n_ommatidia: Number of ommatidia.
        t_pre: Pre-stimulus duration.
        t_post: Post-stimulus duration.
        intensity: Stimulus intensity.
        mode: Stimulus mode.
        device: Torch device for computations.

    Attributes:
        arg_df: DataFrame containing stimulus parameters.
        dt: Time step.
        dots: Instance of the Dots class.
        impulse_durations: List of impulse durations.
        config: Configuration namespace.
        params: List of stimulus parameters.
    """

    arg_df: Optional[pd.DataFrame] = None
    dt: Optional[float] = None

    def __init__(
        self,
        impulse_durations: List[float] = [5e-3, 20e-3],
        max_extent: int = 4,
        dot_column_radius: int = 0,
        bg_intensity: float = 0.5,
        t_stim: float = 5,
        dt: float = 0.005,
        n_ommatidia: int = 721,
        t_pre: float = 2.0,
        t_post: float = 0,
        intensity: float = 1,
        mode: str = "impulse",
        device: torch.device = flyvis.device,
    ):
        self.dots = Dots(
            dot_column_radius=dot_column_radius,
            max_extent=max_extent,
            bg_intensity=bg_intensity,
            t_stim=t_stim,
            dt=dt,
            n_ommatidia=n_ommatidia,
            t_pre=t_pre,
            t_post=t_post,
            intensity=intensity,
            mode=mode,
            device=device,
        )
        self.dt = dt
        self.impulse_durations = impulse_durations

        self.config = self.dots.config
        self.config.update(impulse_durations=impulse_durations)

        self.params = [
            (*p[0], p[1])
            for p in product(self.dots.arg_df.values.tolist(), impulse_durations)
        ]
        self.arg_df = pd.DataFrame(
            self.params,
            columns=[
                "u",
                "v",
                "offset",
                "coordinate_index",
                "intensity",
                "t_impulse",
            ],
        )

    def _params(self, key: int) -> np.ndarray:
        """Get parameters for a specific key.

        Args:
            key: Index of the parameters to retrieve.

        Returns:
            Array of parameters for the given key.
        """
        return self.arg_df.iloc[key].values

    def get_item(self, key: int) -> torch.Tensor:
        """Get a stimulus item for a specific key.

        Args:
            key: Index of the item to retrieve.

        Returns:
            Tensor representing the stimulus sequence.
        """
        u, v, offset, coordinate_index, intensity, t_impulse = self._params(key)
        self.dots.t_impulse = t_impulse
        return self.dots[self.dots.get_stimulus_index(u, v, intensity)]

    def __repr__(self) -> str:
        """Get string representation of the dataset."""
        return repr(self.arg_df)



"""Rendering utils"""

from typing import Literal, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn.functional as F
from PIL import Image

import flyvis


def median(
    x: torch.Tensor,
    kernel_size: int,
    stride: int = 1,
    n_chunks: int = 10,
) -> torch.Tensor:
    """
    Apply median image filter with reflected padding.

    Args:
        x: Input array or tensor of shape (n_samples, n_frames, height, width).
           First and second dimensions are optional.
        kernel_size: Size of the filter kernel.
        stride: Stride for the filter operation.
        n_chunks: Number of chunks to process the data if memory is limited.
            Recursively increases the chunk size until the data fits in memory.

    Returns:
        Filtered array or tensor of the same shape as input.

    Note:
        On GPU, this creates a tensor of kernel_size ** 2 * prod(x.shape) elements,
        consuming significant memory (e.g., ~14 GB for 50 frames of 436x1024 with
        kernel_size 13). In case of a RuntimeError due to memory, the method
        processes the data in chunks.
    """
    # Get padding so that the resulting tensor is of the same shape.
    p = max(kernel_size - 1, 0)
    p_floor = p // 2
    p_ceil = p - p_floor
    padding = (p_floor, p_ceil, p_floor, p_ceil)

    shape = x.shape

    try:
        with torch.no_grad():
            if len(shape) == 2:
                x.unsqueeze_(0).unsqueeze_(0)
            elif len(shape) == 3:
                x.unsqueeze_(0)
            elif len(shape) == 4:
                pass
            else:
                raise ValueError(f"Invalid shape: {shape}")
            assert len(x.shape) == 4
            _x = F.pad(x, padding, mode="reflect")
            _x = _x.unfold(dimension=2, size=kernel_size, step=stride).unfold(
                dimension=3, size=kernel_size, step=stride
            )
            _x = _x.contiguous().view(shape[:4] + (-1,)).median(dim=-1)[0]
            return _x.view(shape)
    except RuntimeError as e:
        if "memory" not in str(e):
            raise e
        if "CUDA" in str(e):
            torch.cuda.empty_cache()
        _x = x.reshape(-1, *x.shape[-2:])
        chunks = torch.chunk(_x, max(n_chunks, 1), dim=0)

        def map_fn(z):
            return median(z, kernel_size, n_chunks=n_chunks - 1)

        _x = torch.cat(tuple(map(map_fn, chunks)), dim=0)
        return _x.view(shape)


def split(
    array: Union[np.ndarray, torch.Tensor],
    out_nelements: int,
    n_splits: int,
    center_crop_fraction: Optional[float] = 0.7,
) -> Union[np.ndarray, torch.Tensor]:
    """
    Split an array into overlapping segments along the last dimension.

    Args:
        array: Input array of shape (..., nelements).
        out_nelements: Number of elements in each output split.
        n_splits: Number of splits to create.
        center_crop_fraction: If not None, the array is centrally cropped in the
            last dimension to this fraction before splitting.

    Returns:
        A new array of shape (n_splits, ..., out_nelements) containing the splits.

    Raises:
        ValueError: If n_splits is less than 0.
        TypeError: If the input array is neither a numpy array nor a torch tensor.

    Note:
        - If n_splits is 1, the entire array is returned (with an added dimension).
        - If n_splits is None or 0, the original array is returned unchanged.
        - Splits may overlap if out_nelements * n_splits > array.shape[-1].
    """
    assert isinstance(array, (np.ndarray, torch.Tensor))
    if center_crop_fraction is not None:
        return split(
            center_crop(array, center_crop_fraction),
            out_nelements,
            n_splits,
            center_crop_fraction=None,
        )

    actual_nelements = array.shape[-1]
    out_nelements = int(out_nelements)

    def take(
        arr: Union[np.ndarray, torch.Tensor], start: int, stop: int
    ) -> Union[np.ndarray, torch.Tensor]:
        if isinstance(arr, np.ndarray):
            return np.take(arr, np.arange(start, stop), axis=-1)[None]
        elif isinstance(arr, torch.Tensor):
            return torch.index_select(arr, dim=-1, index=torch.arange(start, stop))[None]

    if n_splits == 1:
        out = (array[None, :],)
    elif n_splits > 1:
        out = ()
        out_nelements = max(out_nelements, int(actual_nelements / n_splits))
        overlap = np.ceil(
            (out_nelements * n_splits - actual_nelements) / (n_splits - 1)
        ).astype(int)
        for i in range(n_splits):
            start = i * out_nelements - i * overlap
            stop = (i + 1) * out_nelements - i * overlap
            out += (take(array, start, stop),)
    elif n_splits is None or n_splits == 0:
        return array
    else:
        raise ValueError("n_splits must be a non-negative integer or None")

    if isinstance(array, np.ndarray):
        return np.concatenate(out, axis=0)
    elif isinstance(array, torch.Tensor):
        return torch.cat(out, dim=0)


def center_crop(
    array: Union[np.ndarray, torch.Tensor], out_nelements_ratio: float
) -> Union[np.ndarray, torch.Tensor]:
    """
    Centrally crop an array along the last dimension with given ratio.

    Args:
        array: Array of shape (..., nelements).
        out_nelements_ratio: Ratio of output elements to input elements.

    Returns:
        Cropped array of shape (..., out_nelements).
    """

    def take(arr, start, stop):
        if isinstance(arr, np.ndarray):
            return np.take(arr, np.arange(start, stop), axis=-1)
        elif isinstance(arr, torch.Tensor):
            return torch.index_select(arr, dim=-1, index=torch.arange(start, stop))

    nelements = array.shape[-1]
    out_nelements = int(out_nelements_ratio * nelements)
    return take(array, (nelements - out_nelements) // 2, (nelements + out_nelements) // 2)


def hex_center_coordinates(
    n_hex_area: int, img_width: int, img_height: int, center: bool = True
) -> Tuple[np.ndarray, np.ndarray, Tuple[float, float]]:
    """
    Calculate hexagon center coordinates for a given area.

    Args:
        n_hex_area: Number of hexagons in the area.
        img_width: Width of the image.
        img_height: Height of the image.
        center: If True, center the hexagon grid in the image.

    Returns:
        Tuple containing x coordinates, y coordinates, and (dist_w, dist_h).
    """
    # Horizontal extent of the grid
    n = np.floor(np.sqrt(n_hex_area / 3)).astype("int")

    dist_h = img_height / (2 * n + 1)
    dist_w = img_width / (2 * n + 1)

    xs = []
    ys = []
    for q in range(-n, n + 1):
        for r in range(max(-n, -n - q), min(n, n - q) + 1):
            xs.append(dist_w * r)
            ys.append(
                dist_h * (q + r / 2)
            )  # either must be negative or origin must be upper
    xs, ys = np.array(xs), np.array(ys)
    if center:
        xs += img_width // 2
        ys += img_height // 2
    return xs, ys, (dist_w, dist_h)


def is_inside_hex(
    x: torch.Tensor,
    y: torch.Tensor,
    x_centers: torch.Tensor,
    y_centers: torch.Tensor,
    dist_to_edge: float,
    tilt: Union[float, torch.Tensor],
    device: torch.device = flyvis.device,
    dtype: torch.dtype = torch.float16,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Find whether given points are inside the given hexagons.

    Args:
        x: Cartesian x-coordinates of the points (n_points).
        y: Cartesian y-coordinates of the points (n_points).
        x_centers: Cartesian x-centers of the hexagons (n_hexagons).
        y_centers: Cartesian y-centers of the hexagons (n_hexagons).
        dist_to_edge: Euclidean distance from center to edge of the hexagon.
        tilt: Angle of hexagon counter-clockwise tilt in radians.
        device: Torch device to use for computations.
        dtype: Data type for torch tensors.

    Returns:
        Tuple containing:
        - vertices: Cartesian coordinates of the hexagons' vertices (7, 2, n_hexagons).
        - is_inside: Boolean tensor indicating whether points are inside
            (n_points, n_hexagons).

    Info: Credits
        Adapted from Roman Vaxenburg's original implementation.
    """
    if not isinstance(tilt, torch.Tensor):
        tilt = torch.tensor(tilt, device=device)

    R = torch.tensor(
        [
            [torch.cos(tilt), -torch.sin(tilt)],
            [torch.sin(tilt), torch.cos(tilt)],
        ],
        dtype=dtype,
        device=device,
    )  # rotation matrix
    pi = torch.tensor(np.pi, device=device, dtype=dtype)
    R60 = torch.tensor(
        [
            [torch.cos(pi / 3), -torch.sin(pi / 3)],
            [torch.sin(pi / 3), torch.cos(pi / 3)],
        ],
        dtype=dtype,
        device=device,
    )  # rotation matrix

    # Generate hexagon vertices
    dist_to_vertex = 2 / np.sqrt(3) * dist_to_edge
    vertices = torch.zeros(7, 2, dtype=dtype, device=device)
    vertices[0, :] = torch.matmul(
        R, torch.tensor([dist_to_vertex, 0], dtype=dtype, device=device)
    )
    for i in range(1, 7):
        vertices[i] = torch.matmul(R60, vertices[i - 1])
    vertices = vertices[:, :, None]
    vertices = torch.cat(
        (
            vertices[:, 0:1, :] + x_centers[None, None, :],
            vertices[:, 1:2, :] + y_centers[None, None, :],
        ),
        dim=1,
    )  # (7, 2, n_hexagons)

    # Generate is_inside output
    is_inside = torch.ones(len(x), len(x_centers), dtype=torch.bool, device=device)
    for i in range(6):
        x1, y1 = vertices[i, :, :]  # x1, y1: (n_hexagons)
        x2, y2 = vertices[i + 1, :, :]
        slope = (y2 - y1) / (x2 - x1)  # (n_hexagons)
        f_center = y1 + slope * (x_centers - x1) - y_centers  # (n_hexagons)
        f_points = (
            y1[None, :] + slope[None, :] * (x[:, None] - x1[None, :]) - y[:, None]
        )  # (n_points, n_hexagons)
        is_inside = torch.logical_and(is_inside, f_center.sign() == f_points.sign())

    return vertices, is_inside  # (7, 2, n_hexagons), (n_points, n_hexagons)


def render_bars_cartesian(
    img_height_px: int,
    img_width_px: int,
    bar_width_px: int,
    bar_height_px: int,
    bar_loc_horizontal_px: int,
    bar_loc_vertical_px: int,
    n_bars: int,
    bar_intensity: float,
    bg_intensity: float,
    rotate: float = 0,
) -> np.ndarray:
    """
    Render bars in a cartesian coordinate system.

    Args:
        img_height_px: Height of the image in pixels.
        img_width_px: Width of the image in pixels.
        bar_width_px: Width of each bar in pixels.
        bar_height_px: Height of each bar in pixels.
        bar_loc_horizontal_px: Horizontal location of the bars in pixels.
        bar_loc_vertical_px: Vertical location of the bars in pixels.
        n_bars: Number of bars to generate.
        bar_intensity: Intensity of the bars.
        bg_intensity: Intensity of the background.
        rotate: Rotation angle in degrees.

    Returns:
        Generated image as a numpy array.
    """
    bar_spacing = int(img_width_px / n_bars - bar_width_px)

    height_slice = slice(
        int(bar_loc_vertical_px - bar_height_px / 2),
        int(bar_loc_vertical_px + bar_height_px / 2) + 1,
    )

    img = np.ones([img_height_px, img_width_px]) * bg_intensity

    loc_w = int(bar_loc_horizontal_px - bar_width_px / 2)
    for i in range(n_bars):
        #  Fill background with bars.
        start = max(loc_w + i * bar_width_px + i * bar_spacing, 0)
        width_slice = slice(start, loc_w + (i + 1) * bar_width_px + i * bar_spacing + 1)
        img[height_slice, width_slice] = bar_intensity

    if rotate % 360 != 0:
        img = rotate_image(img, angle=rotate)

    return img


def render_gratings_cartesian(
    img_height_px: int,
    img_width_px: int,
    spatial_period_px: float,
    grating_intensity: float,
    bg_intensity: float,
    grating_height_px: Optional[int] = None,
    grating_width_px: Optional[int] = None,
    grating_phase_px: float = 0,
    rotate: float = 0,
) -> np.ndarray:
    """
    Render gratings in a cartesian coordinate system.

    Args:
        img_height_px: Height of the image in pixels.
        img_width_px: Width of the image in pixels.
        spatial_period_px: Spatial period of the gratings in pixels.
        grating_intensity: Intensity of the gratings.
        bg_intensity: Intensity of the background.
        grating_height_px: Height of the grating area in pixels.
        grating_width_px: Width of the grating area in pixels.
        grating_phase_px: Phase of the gratings in pixels.
        rotate: Rotation angle in degrees.

    Returns:
        Generated image as a numpy array.
    """
    # to save time at library import
    from scipy.signal import square

    t = (
        2
        * np.pi
        / (spatial_period_px / img_width_px)
        * (
            np.linspace(-1 / 2, 1 / 2, int(img_width_px))
            - grating_phase_px / img_width_px
        )
    )

    gratings = np.tile(square(t), img_height_px).reshape(img_height_px, img_width_px)
    gratings[gratings == -1] = bg_intensity
    gratings[gratings == 1] = grating_intensity

    if grating_height_px:
        mask = np.ones_like(gratings).astype(bool)

        height_slice = slice(
            int(img_height_px // 2 - grating_height_px / 2),
            int(img_height_px // 2 + grating_height_px / 2) + 1,
        )
        mask[height_slice] = False
        gratings[mask] = 0.5

    if grating_width_px:
        mask = np.ones_like(gratings).astype(bool)

        width_slice = slice(
            int(img_width_px // 2 - grating_width_px / 2),
            int(img_width_px // 2 + grating_width_px / 2) + 1,
        )
        mask[:, width_slice] = False
        gratings[mask] = 0.5

    if rotate % 360 != 0:
        gratings = rotate_image(gratings, angle=rotate)

    return gratings


def rotate_image(img: np.ndarray, angle: float = 0) -> np.ndarray:
    """
    Rotate an image by a given angle.

    Args:
        img: Input image as a numpy array.
        angle: Rotation angle in degrees.

    Returns:
        Rotated image as a numpy array.
    """
    h, w = img.shape

    diagonal = int(np.sqrt(h**2 + w**2))

    pad_in_height = (diagonal - h) // 2
    pad_in_width = (diagonal - w) // 2

    img = np.pad(
        img,
        ((pad_in_height, pad_in_height), (pad_in_width, pad_in_width)),
        mode="edge",
    )

    img = Image.fromarray((255 * img).astype("uint8")).rotate(
        angle, Image.BILINEAR, False, None
    )
    img = np.array(img, dtype=float) / 255.0

    padded_h, padded_w = img.shape
    return img[
        pad_in_height : padded_h - pad_in_height,
        pad_in_width : padded_w - pad_in_width,
    ]


def resample(
    stims: torch.Tensor,
    t_stim: float,
    dt: float,
    dim: int = 0,
    device: torch.device = flyvis.device,
    return_indices: bool = False,
) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
    """
    Resample a set of stimuli for a given stimulus duration and time step.

    Args:
        stims: Stimuli tensor of shape (#conditions, #hexals).
        t_stim: Stimulus duration in seconds.
        dt: Integration time constant in seconds.
        dim: Dimension along which to resample.
        device: Torch device to use for computations.
        return_indices: If True, return the indices used for resampling.

    Returns:
        Resampled stimuli tensor of shape (#frames, #hexals), or a tuple of
        (resampled stimuli, indices) if return_indices is True.
    """
    n_offsets = stims.shape[dim]
    # round to nearest integer
    # this results in unequal counts of each frame usually by +-1
    indices = torch.linspace(0, n_offsets - 1, int(t_stim / dt), device=device).long()
    if not return_indices:
        return torch.index_select(stims, dim, indices)
    return torch.index_select(stims, dim, indices), indices


def shuffle(
    stims: torch.Tensor, randomstate: Optional[np.random.RandomState] = None
) -> torch.Tensor:
    """
    Randomly shuffle stimuli along the frame dimension.

    Args:
        stims: Stimuli tensor of shape (N (optional), #frames, #hexals).
        randomstate: Random state for reproducibility.

    Returns:
        Shuffled stimuli tensor.
    """
    if len(stims.shape) == 3:
        # assume (smples frames hexals)
        def _shuffle(x):
            return shuffle(x, randomstate)

        return torch.stack(list(map(_shuffle, stims)), dim=0)
    perms = (
        randomstate.permutation(stims.shape[0])
        if randomstate is not None
        else np.random.permutation(stims.shape[0])
    )
    return stims[perms]


def resample_grating(
    grating: torch.Tensor, t_stim: float, dt: float, temporal_frequency: float
) -> torch.Tensor:
    """
    Resample a grating stimulus for a given duration and temporal frequency.

    Args:
        grating: Input grating tensor.
        t_stim: Stimulus duration in seconds.
        dt: Time step in seconds.
        temporal_frequency: Temporal frequency of the grating.

    Returns:
        Resampled grating tensor.
    """
    n_frames = int(t_stim / dt)
    t_period = 1 / temporal_frequency
    _grating = resample(grating, t_period, dt)
    _grating = _grating.repeat(np.ceil(n_frames / _grating.shape[0]).astype(int), 1)
    return _grating[:n_frames]


def pad(
    stim: torch.Tensor,
    t_stim: float,
    dt: float,
    fill: float = 0,
    mode: Literal["end", "start"] = "end",
    pad_mode: Literal["value", "continue", "reflect"] = "value",
) -> torch.Tensor:
    """
    Pad the second to last dimension of a stimulus tensor.

    Args:
        stim: Stimulus tensor of shape (..., n_frames, n_hexals).
        t_stim: Target stimulus duration in seconds.
        dt: Integration time constant in seconds.
        fill: Value to fill with if pad_mode is "value".
        mode: Padding mode, either "end" or "start".
        pad_mode: Padding type, either "value", "continue", or "reflect".

    Returns:
        Padded stimulus tensor.
    """
    diff = int(t_stim / dt) - stim.shape[-2]
    if diff <= 0:
        return stim

    # Pad the second-to-last dimension (n_frames)
    # Format: (pad_last_dim_left, pad_last_dim_right,
    #          pad_second_to_last_dim_before, pad_second_to_last_dim_after)
    if mode == "end":
        pad = (0, 0, 0, diff)  # Pad after the existing frames
    elif mode == "start":
        pad = (0, 0, diff, 0)  # Pad before the existing frames

    if pad_mode == "value":
        return torch.nn.functional.pad(stim, pad=pad, mode="constant", value=fill)
    elif pad_mode == "continue":
        return repeat_last(stim, -2, diff)
    else:
        return torch.nn.functional.pad(stim, pad=pad, mode=pad_mode)


def repeat_last(stim: torch.Tensor, dim: int, n_repeats: int) -> torch.Tensor:
    """
    Repeat the last frame of a stimulus tensor along a specified dimension.

    Args:
        stim: Input stimulus tensor.
        dim: Dimension along which to repeat.
        n_repeats: Number of times to repeat the last frame.

    Returns:
        Stimulus tensor with the last frame repeated.
    """
    last = stim.index_select(dim, torch.tensor([stim.size(dim) - 1], device=stim.device))
    stim = torch.cat((stim, last.repeat_interleave(n_repeats, dim=dim)), dim=dim)
    return stim



import logging
from itertools import product
from numbers import Number
from typing import Iterable, Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from datamate import Directory, Namespace, root
from matplotlib.colors import Colormap
from matplotlib.patches import RegularPolygon
from tqdm.auto import tqdm
from typing_extensions import Literal

import flyvis
from flyvis import renderings_dir
from flyvis.analysis.visualization.plots import quick_hex_scatter
from flyvis.analysis.visualization.plt_utils import init_plot
from flyvis.datasets.datasets import StimulusDataset

from .rendering import HexEye
from .rendering.utils import pad, resample, shuffle

logging = logging.getLogger(__name__)

__all__ = ["RenderedOffsets", "MovingBar", "MovingEdge"]


@root(renderings_dir)
class RenderedOffsets(Directory):
    """Rendered offsets for the moving bar stimulus.

    This class precomputes the offsets for moving bar (edge) stimuli and stores them
    in a directory. At runtime, the offsets are resampled to efficiently generate
    stimuli with different durations and temporal resolutions.

    Args:
        offsets: List of offset values.
        angles: List of angle values in degrees.
        widths: List of width values.
        intensities: List of intensity values.
        led_width: Width of LED in radians.
        height: Height of the bar in radians.
        n_bars: Number of bars.
        bg_intensity: Background intensity.
        bar_loc_horizontal: Horizontal location of the bar in radians.

    Attributes:
        offsets (ArrayFile): Rendered offsets for different stimulus parameters.
    """

    def __init__(
        self,
        offsets: list[int] = list(range(-10, 11)),
        angles: list[int] = [0, 30, 60, 90, 120, 150, 180, 210, 240, 270, 300, 330],
        widths: list[int] = [1, 2, 4],
        intensities: list[int] = [0, 1],
        led_width: float = np.radians(2.25),
        height: float = np.radians(2.25) * 9,
        n_bars: int = 1,
        bg_intensity: float = 0.5,
        bar_loc_horizontal: float = np.radians(90),
    ):
        eye = HexEye(721, 25)

        params = list(product(angles, widths, intensities))

        sequences = {}
        _tqdm = tqdm(total=len(params), desc="building stimuli")
        for angle in angles:
            for width in widths:
                for intensity in intensities:
                    offset_bars = eye.render_offset_bars(
                        bar_width_rad=width * led_width,
                        bar_height_rad=height,
                        n_bars=n_bars,
                        offsets=np.array(offsets) * led_width,
                        bar_intensity=intensity,
                        bg_intensity=bg_intensity,
                        moving_angle=angle,
                        bar_loc_horizontal=bar_loc_horizontal,
                    )
                    sequences[(angle, width, intensity)] = offset_bars
                    _tqdm.update()

        _tqdm.close()
        self.offsets = torch.stack([sequences[p] for p in params]).cpu().numpy()


class MovingBar(StimulusDataset):
    """Moving bar stimulus.

    Args:
        widths: Width of the bar in half ommatidia.
        offsets: First and last offset to the central column in half ommatidia.
        intensities: Intensity of the bar.
        speeds: Speed of the bar in half ommatidia per second.
        height: Height of the bar in half ommatidia.
        dt: Time step in seconds.
        device: Device to store the stimulus.
        bar_loc_horizontal: Horizontal location of the bar in radians from left to
            right of image plane. np.radians(90) is the center.
        post_pad_mode: Padding mode after the stimulus. One of 'continue', 'value',
            'reflect'. If 'value' the padding is filled with `bg_intensity`.
        t_pre: Time before the stimulus in seconds.
        t_post: Time after the stimulus in seconds.
        build_stim_on_init: Build the stimulus on initialization.
        shuffle_offsets: Shuffle the offsets to remove spatio-temporal correlation.
        seed: Seed for the random state.
        angles: List of angles in degrees.

    Attributes:
        config (Namespace): Configuration parameters.
        omm_width (float): Width of ommatidium in radians.
        led_width (float): Width of LED in radians.
        angles (np.ndarray): Array of angles in degrees.
        widths (np.ndarray): Array of widths in half ommatidia.
        offsets (np.ndarray): Array of offsets in half ommatidia.
        intensities (np.ndarray): Array of intensities.
        speeds (np.ndarray): Array of speeds in half ommatidia per second.
        bg_intensity (float): Background intensity.
        n_bars (int): Number of bars.
        bar_loc_horizontal (float): Horizontal location of bar in radians.
        t_stim (np.ndarray): Stimulation times for each speed.
        t_stim_max (float): Maximum stimulation time.
        height (float): Height of bar in radians.
        post_pad_mode (str): Padding mode after the stimulus.
        arg_df (pd.DataFrame): DataFrame of stimulus parameters.
        arg_group_df (pd.DataFrame): Grouped DataFrame of stimulus parameters.
        device (str): Device for storing stimuli.
        shuffle_offsets (bool): Whether to shuffle offsets.
        randomstate (np.random.RandomState): Random state for shuffling.
    """

    arg_df: pd.DataFrame = None

    def __init__(
        self,
        widths: list[int] = [1, 2, 4],
        offsets: tuple[int, int] = (-10, 11),
        intensities: list[float] = [0, 1],
        speeds: list[float] = [2.4, 4.8, 9.7, 13, 19, 25],
        height: int = 9,
        dt: float = 1 / 200,
        device: str = flyvis.device,
        bar_loc_horizontal: float = np.radians(90),
        post_pad_mode: Literal["continue", "value", "reflect"] = "value",
        t_pre: float = 1.0,
        t_post: float = 1.0,
        build_stim_on_init: bool = True,
        shuffle_offsets: bool = False,
        seed: int = 0,
        angles: list[int] = [0, 30, 60, 90, 120, 150, 180, 210, 240, 270, 300, 330],
    ) -> None:
        super().__init__()
        # HexEye parameter
        self.omm_width = np.radians(5.8)  # radians(5.8 degree)

        # Monitor parameter
        self.led_width = np.radians(2.25)  # Gruntman et al. 2018

        _locals = locals()
        self.config = Namespace({
            arg: _locals[arg]
            for arg in [
                "widths",
                "offsets",
                "intensities",
                "speeds",
                "height",
                "bar_loc_horizontal",
                "shuffle_offsets",
                "post_pad_mode",
                "t_pre",
                "t_post",
                "dt",
                "angles",
            ]
        })

        # Stim parameter
        self.angles = np.array(angles)
        self.widths = np.array(widths)  # half ommatidia
        if len(offsets) == 2:
            self.offsets = np.arange(*offsets)  # half ommatidia
        else:
            assert (
                np.mean(offsets[1:] - offsets[:-1]) == 1
            )  # t_stim assumes spacing of 1 corresponding to 2.25 deg
            self.offsets = offsets
        self.intensities = np.array(intensities)
        self.speeds = np.array(speeds)
        self.bg_intensity = 0.5
        self.n_bars = 1
        self.bar_loc_horizontal = bar_loc_horizontal

        self.t_stim = (len(self.offsets) * self.led_width) / (
            self.speeds * self.omm_width
        )
        self.t_stim_max = np.max(self.t_stim)

        self._speed_to_t_stim = dict(zip(self.speeds, self.t_stim))

        self.height = self.led_width * height

        self.post_pad_mode = post_pad_mode
        self._t_pre = t_pre
        self._t_post = t_post

        params = [
            (*p[:-1], *p[-1])
            for p in list(
                product(
                    self.angles,
                    self.widths,
                    self.intensities,
                    zip(self.t_stim, self.speeds),
                )
            )
        ]
        self.arg_df = pd.DataFrame(
            params, columns=["angle", "width", "intensity", "t_stim", "speed"]
        )

        self.arg_group_df = self.arg_df.groupby(
            ["angle", "width", "intensity"], sort=False, as_index=False
        ).all()

        self.device = device
        self.shuffle_offsets = shuffle_offsets
        self.randomstate = None
        if self.shuffle_offsets:
            self.randomstate = np.random.RandomState(seed=seed)

        self._dt = dt

        self._built = False
        if build_stim_on_init:
            self._build()
            self._resample()
            self._built = True

    @property
    def dt(self) -> float:
        """Time step in seconds."""
        return getattr(self, "_dt", None)

    @dt.setter
    def dt(self, value: float) -> None:
        if self._dt == value:
            self._dt = value
            if self._built:
                self._resample()
            return
        logging.warning(
            "Cannot override dt=%s because responses with dt=%s are initialized. "
            "Keeping dt=%s.",
            value,
            self._dt,
            self._dt,
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}\n"
            + "Config:\n"
            + repr(self.config)
            + "Stimulus parameter:\n"
            + repr(self.arg_df)
        )

    @property
    def t_pre(self) -> float:
        """Time before stimulus onset in seconds."""
        return self._t_pre

    @property
    def t_post(self) -> float:
        """Time after stimulus offset in seconds."""
        return self._t_post

    def _build(self) -> None:
        """Build the stimulus."""
        self.wrap = RenderedOffsets(
            dict(
                angles=self.angles,
                widths=self.widths,
                intensities=self.intensities,
                offsets=self.offsets,
                led_width=self.led_width,
                height=self.height,
                n_bars=self.n_bars,
                bg_intensity=self.bg_intensity,
                bar_loc_horizontal=self.bar_loc_horizontal,
            )
        )

        self._offsets = torch.tensor(self.wrap.offsets[:], device=self.device)
        self._built = True

    def _resample(self) -> None:
        """Resample the stimulus at runtime."""
        # resampling at runtime to allow for changing dt and to save GB of
        # storage.
        self.sequences = {}
        self.indices = {}
        for t, speed in zip(self.t_stim, self.speeds):
            sequence, indices = resample(
                self._offsets,
                t,
                self.dt,
                dim=1,
                device=self.device,
                return_indices=True,
            )
            if self.shuffle_offsets:
                # breakpoint()
                sequence = shuffle(sequence, self.randomstate)
            sequence = pad(
                sequence,
                t + self.t_pre,
                self.dt,
                mode="start",
                fill=self.bg_intensity,
            )
            sequence = pad(
                sequence,
                t + self.t_pre + self.t_post,
                self.dt,
                mode="end",
                pad_mode=self.post_pad_mode,
                fill=self.bg_intensity,
            )
            # Because we fix the distance that the bar moves but vary speeds we
            # have different stimulation times. To make all sequences equal
            # length for storing them in a single tensor, we pad them with nans
            # based on the maximal stimulation time (slowest speed). The nans
            # can later be removed before processing the traces.
            sequence = pad(
                sequence,
                self.t_stim_max + self.t_pre + self.t_post,
                self.dt,
                mode="end",
                fill=np.nan,
            )
            self.sequences[speed] = sequence
            self.indices[speed] = indices

    def _key(self, angle: float, width: float, intensity: float, speed: float) -> int:
        """Get the key for a specific stimulus configuration."""
        try:
            return self.arg_df.query(
                f"angle=={angle}"
                f" & width=={width}"
                f" & intensity == {intensity}"
                f" & speed == {speed}"
            ).index.values.item()
        except ValueError:
            raise ValueError(
                f"angle: {angle}, width: {width}, intensity: {intensity}, "
                f"speed: {speed} invalid."
            ) from None

    def get_sequence_id_from_arguments(
        self, angle: float, width: float, intensity: float, speed: float
    ) -> int:
        """Get sequence ID from stimulus arguments."""
        return self.get_stimulus_index(locals())

    def _params(self, key: int) -> np.ndarray:
        """Get parameters for a given key."""
        return self.arg_df.iloc[key].values

    def _group_key(self, angle: float, width: float, intensity: float) -> int:
        """Get group key for a specific stimulus configuration."""
        return self.arg_group_df.query(
            f"angle=={angle}" f" & width=={width}" f" & intensity == {intensity}"
        ).index.values.item()

    def _group_params(self, key: int) -> np.ndarray:
        """Get group parameters for a given key."""
        return self.arg_group_df.iloc[key].values

    def get(
        self, angle: float, width: float, intensity: float, speed: float
    ) -> torch.Tensor:
        """Get stimulus for specific parameters."""
        key = self._key(angle, width, intensity, speed)
        return self[key]

    def get_item(self, key: int) -> torch.Tensor:
        """Get stimulus for a specific key."""
        angle, width, intensity, _, speed = self._params(key)
        return self.sequences[speed][self._group_key(angle, width, intensity)]

    def mask(
        self,
        angle: Optional[float] = None,
        width: Optional[float] = None,
        intensity: Optional[float] = None,
        speed: Optional[float] = None,
        t_stim: Optional[float] = None,
    ) -> np.ndarray:
        """Create a mask for specific stimulus parameters."""
        # 22x faster than df.query
        values = self.arg_df.values

        def iterparam(param, name, axis, and_condition):
            condition = np.zeros(len(values)).astype(bool)
            if isinstance(param, Iterable):
                for p in param:
                    _new = values.take(axis, axis=1) == p
                    assert any(_new), f"{name} {p} not in dataset."
                    condition = np.logical_or(condition, _new)
            else:
                _new = values.take(axis, axis=1) == param
                assert any(_new), f"{name} {param} not in dataset."
                condition = np.logical_or(condition, _new)
            return condition & and_condition

        condition = np.ones(len(values)).astype(bool)
        if angle is not None:
            condition = iterparam(angle, "angle", 0, condition)
        if width is not None:
            condition = iterparam(width, "width", 1, condition)
        if intensity is not None:
            condition = iterparam(intensity, "intensity", 2, condition)
        if t_stim is not None:
            condition = iterparam(t_stim, "t_stim", 3, condition)
        if speed is not None:
            condition = iterparam(speed, "speed", 4, condition)
        return condition

    @property
    def time(self) -> np.ndarray:
        """Time array for the stimulus."""
        return np.arange(-self.t_pre, self.t_stim_max + self.t_post - self.dt, self.dt)

    def stimulus(
        self,
        angle: Optional[float] = None,
        width: Optional[float] = None,
        intensity: Optional[float] = None,
        speed: Optional[float] = None,
        pre_stim: bool = True,
        post_stim: bool = True,
    ) -> np.ndarray:
        """Get stimulus for specific parameters.

        Args:
            angle: Angle of the bar.
            width: Width of the bar.
            intensity: Intensity of the bar.
            speed: Speed of the bar.
            pre_stim: Include pre-stimulus period.
            post_stim: Include post-stimulus period.

        Returns:
            Stimulus array.
        """
        key = self._key(angle, width, intensity, speed)
        stim = self[key][:, 360].cpu().numpy()
        if not post_stim:
            stim = filter_post([stim], self.t_post, self.dt).squeeze()
        if not pre_stim:
            stim = filter_pre(stim[None], self.t_pre, self.dt).squeeze()
        return stim

    def stimulus_parameters(
        self,
        angle: Optional[float] = None,
        width: Optional[float] = None,
        intensity: Optional[float] = None,
        speed: Optional[float] = None,
    ) -> tuple[list, ...]:
        """Get stimulus parameters."""

        def _number_to_list(*args):
            returns = tuple()
            for arg in args:
                if isinstance(arg, Number):
                    returns += ([arg],)
                else:
                    returns += (arg,)
            return returns

        angle, width, speed, intensity = _number_to_list(angle, width, speed, intensity)
        angle = angle or self.angles
        width = width or self.widths
        intensity = intensity or self.intensities
        speed = speed or self.speeds
        return angle, width, intensity, speed

    def sample_shape(
        self,
        angle: Optional[float] = None,
        width: Optional[float] = None,
        intensity: Optional[float] = None,
        speed: Optional[float] = None,
    ) -> tuple[int, ...]:
        """Get shape of stimulus sample for given parameters."""
        if isinstance(angle, Number):
            angle = [angle]
        if isinstance(width, Number):
            width = [width]
        if isinstance(speed, Number):
            speed = [speed]
        if isinstance(intensity, Number):
            intensity = [intensity]
        angle = angle or self.angles
        width = width or self.widths
        intensity = intensity or self.intensities
        speed = speed or self.speeds
        return (
            len(angle),
            len(width),
            len(intensity),
            len(speed),
        )

    def time_to_center(self, speed: float) -> float:
        """Calculate time for bar to reach center at given speed."""
        # Note: time = distance / velocity, i.e.
        #     time = (n_leds * led_width) / (speed * omm_width)
        #     with speed in ommatidia / s.
        return np.abs(self.config.offsets[0]) * self.led_width / (speed * self.omm_width)

    def get_time_with_origin_at_onset(self) -> np.ndarray:
        """Get time array with origin at stimulus onset."""
        return np.linspace(
            -self.t_pre,
            self.t_stim_max - self.t_pre + self.t_post,
            int(self.t_stim_max / self.dt)
            + int(self.t_post / self.dt)
            + int(self.t_pre / self.dt),
        )

    def get_time_with_origin_at_center(self, speed: float) -> np.ndarray:
        """Get time array with origin where bar reaches central column."""
        time_to_center = self.time_to_center(speed)
        n_steps = (
            int(self.t_stim_max / self.dt)
            + int(self.t_post / self.dt)
            + int(self.t_pre / self.dt)
        )
        return np.linspace(
            -(self.t_pre + time_to_center),
            n_steps * self.dt - (self.t_pre + time_to_center),
            n_steps,
        )

    def stimulus_cartoon(
        self,
        angle: float,
        width: float,
        intensity: float,
        speed: float,
        time_after_stimulus_onset: float = 0.5,
        fig: Optional[plt.Figure] = None,
        ax: Optional[plt.Axes] = None,
        facecolor: str = "#000000",
        cmap: Colormap = plt.cm.bone,
        alpha: float = 0.5,
        vmin: float = 0,
        vmax: float = 1,
        edgecolor: str = "none",
        central_hex_color: str = "#2f7cb9",
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Create a cartoon representation of the stimulus."""
        fig, ax = init_plot(fig=fig, ax=ax)

        time = (
            np.arange(
                0,
                self.t_pre + self.t_stim_max + self.t_post - self.dt,
                self.dt,
            )
            - self.t_pre
        )
        index = np.argmin(np.abs(time - time_after_stimulus_onset))

        fig, ax, _ = quick_hex_scatter(
            self.get(angle=angle, width=width, speed=speed, intensity=intensity)
            .cpu()
            .numpy()[index],
            vmin=vmin,
            vmax=vmax,
            cbar=False,
            figsize=[1, 1],
            max_extent=5,
            fig=fig,
            ax=ax,
            cmap=cmap,
            alpha=alpha,
            edgecolor=edgecolor,
            **kwargs,
        )
        rotation = np.array([
            [
                np.cos(np.radians(angle - 90)),
                -np.sin(np.radians(angle - 90)),
            ],
            [
                np.sin(np.radians(angle - 90)),
                np.cos(np.radians(angle - 90)),
            ],
        ])
        x = rotation @ np.array([0, -5])
        dx = rotation @ np.array([0, 1])
        ax.arrow(
            *x,
            *dx,
            facecolor=facecolor,
            width=0.75,
            head_length=2.5,
            edgecolor="k",
            linewidth=0.25,
        )
        _hex = RegularPolygon(
            (0, 0),
            numVertices=6,
            radius=1,
            linewidth=1,
            orientation=np.radians(30),
            edgecolor=central_hex_color,
            facecolor=central_hex_color,
            alpha=1,
            ls="-",
        )
        ax.add_patch(_hex)

        return fig, ax


class MovingEdge(MovingBar):
    """Moving edge stimulus.

    This class creates a moving edge stimulus by using a very wide bar.

    Args:
        offsets: First and last offset to the central column in half ommatidia.
        intensities: Intensity of the edge.
        speeds: Speed of the edge in half ommatidia per second.
        height: Height of the edge in half ommatidia.
        dt: Time step in seconds.
        device: Device to store the stimulus.
        post_pad_mode: Padding mode after the stimulus.
        t_pre: Time before the stimulus in seconds.
        t_post: Time after the stimulus in seconds.
        build_stim_on_init: Build the stimulus on initialization.
        shuffle_offsets: Shuffle the offsets to remove spatio-temporal correlation.
        seed: Seed for the random state.
        angles: List of angles in degrees.

    Note:
        This class uses a very wide bar (width=80) under the hood to render an
        edge stimulus.
    """

    def __init__(
        self,
        offsets: tuple[int, int] = (-10, 11),
        intensities: list[float] = [0, 1],
        speeds: list[float] = [2.4, 4.8, 9.7, 13, 19, 25],
        height: int = 9,
        dt: float = 1 / 200,
        device: str = flyvis.device,
        post_pad_mode: Literal["continue", "value", "reflect"] = "continue",
        t_pre: float = 1.0,
        t_post: float = 1.0,
        build_stim_on_init: bool = True,
        shuffle_offsets: bool = False,
        seed: int = 0,
        angles: list[int] = [0, 30, 60, 90, 120, 150, 180, 210, 240, 270, 300, 330],
        **kwargs,
    ) -> None:
        super().__init__(
            widths=[80],
            offsets=offsets,
            intensities=intensities,
            speeds=speeds,
            height=height,
            dt=dt,
            device=device,
            bar_loc_horizontal=np.radians(0),
            post_pad_mode=post_pad_mode,
            t_pre=t_pre,
            t_post=t_post,
            build_stim_on_init=build_stim_on_init,
            shuffle_offsets=shuffle_offsets,
            seed=seed,
            angles=angles,
        )


def filter_post(resp: np.ndarray, t_post: float, dt: float) -> np.ndarray:
    """Remove post-stimulus time from responses.

    Args:
        resp: Response array of shape (n_samples, n_frames, ...).
        t_post: Post-stimulus time in seconds.
        dt: Time step in seconds.

    Returns:
        Filtered response array.

    Note:
        The post stimulus time is per speed in 'first_nan - int(t_post / dt):first_nan'.
        Assuming resp can be partitioned along the temporal dimension as:
        |pre stim|stimulus|post stimulus|nan padding|
        |0:t_pre:|t_pre + t_stim:|t_pre + t_stim + t_post:|t_pre + t_stim_max + t_post|
    """
    _resp = []
    # for each sample
    for r in resp:
        # create a temporal mask
        mask = np.ones(r.shape[0]).astype(bool)
        # check where the nans are in (temporal, <#cell_type>)
        _nans = np.isnan(r)
        # if there are nans
        if _nans.any():
            # find the first nan (but latest across cell types, cause it needs
            # time to propagate)
            _first_nan_index = np.argmax(_nans, axis=0).max()

            # mask out the post stimulus time
            mask[_first_nan_index - int(t_post / dt) : _first_nan_index] = False

        else:
            mask[-int(t_post / dt) :] = False

        _resp.append(r[mask])
    return np.array(_resp)


def filter_pre(resp: np.ndarray, t_pre: float, dt: float) -> np.ndarray:
    """Remove pre-stimulus time from responses.

    Args:
        resp: Response array of shape (n_samples, n_frames, ...).
        t_pre: Pre-stimulus time in seconds.
        dt: Time step in seconds.

    Returns:
        Filtered response array.
    """
    return resp[:, int(t_pre / dt) :]



import tempfile
from pathlib import Path

import numpy as np
import pytest
from PIL import Image

from flyvis import connectome_file
from flyvis.connectome import ConnectomeFromAvgFilters


@pytest.fixture(scope="session")
def connectome(tmp_path_factory):
    return ConnectomeFromAvgFilters(
        tmp_path_factory.mktemp("tmp") / "test",
        dict(file=connectome_file.name, extent=1, n_syn_fill=1),
    )


@pytest.fixture(scope="session")
def sequence_path(tmp_path_factory):
    sequences = np.random.rand(20, 10, 64, 64)
    sequences = np.transpose(sequences, (1, 0, 2, 3)) / 255.0
    path = tmp_path_factory.mktemp("tmp") / "sequences.npy"
    np.save(path, sequences)
    return str(path)


@pytest.fixture(scope="session")
def mock_sintel_data():
    """Create a minimal mock Sintel dataset structure with original dimensions."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)

        # Original dimensions
        HEIGHT, WIDTH = 436, 1024

        for seq_name in ["alley_1", "alley_2"]:
            # Create directory structure
            (tmp_path / "training/final" / seq_name).mkdir(parents=True)
            (tmp_path / "training/flow" / seq_name).mkdir(parents=True)
            (tmp_path / "training/depth" / seq_name).mkdir(parents=True)

            # Create dummy files with original dimensions
            for i in range(5):
                # Luminance (final) - (436, 1024)
                img = (np.random.uniform(0, 1, (HEIGHT, WIDTH)) * 255).astype(np.uint8)
                Image.fromarray(img).save(
                    tmp_path / f"training/final/{seq_name}/frame_{i:04d}.png"
                )

                # Flow - (2, 436, 1024)
                with open(
                    tmp_path / f"training/flow/{seq_name}/frame_{i:04d}.flo", 'wb'
                ) as f:
                    # Write header
                    np.array([202021.25], dtype=np.float32).tofile(f)  # Magic number
                    np.array([WIDTH, HEIGHT], dtype=np.int32).tofile(f)  # Dimensions
                    # Write flow data
                    np.random.randn(HEIGHT, WIDTH, 2).astype(np.float32).tofile(f)

                # Depth - (436, 1024)
                with open(
                    tmp_path / f"training/depth/{seq_name}/frame_{i:04d}.dpt", 'wb'
                ) as f:
                    # Write header
                    np.array([1, WIDTH, HEIGHT], dtype=np.int32).tofile(f)  # Dimensions
                    # Write depth data
                    np.random.randn(HEIGHT, WIDTH).astype(np.float32).tofile(f)

        yield tmp_path



import numpy as np
import pytest
import torch

from flyvis.network.stimulus import Stimulus


@pytest.fixture
def stimulus(connectome) -> Stimulus:
    return Stimulus(connectome)


def test_stimulus_init(stimulus):
    assert stimulus.buffer.shape == (1, 1, stimulus.n_nodes)
    assert stimulus.buffer.sum() == 0.0
    assert hasattr(stimulus, "layer_index")
    assert hasattr(stimulus, "central_cells_index")
    assert hasattr(stimulus, "input_index")


def test_add_input(stimulus: Stimulus):
    # single batch, single frame
    batch_size = 1
    frames = 1
    x = torch.ones((batch_size, frames, 1, stimulus.n_input_elements))
    stimulus.add_input(x)
    assert stimulus.buffer.shape == (batch_size, frames, stimulus.n_nodes)
    assert stimulus.buffer.sum() == batch_size * frames * np.prod(
        stimulus.input_index.shape
    )

    # multi batch, multi frame
    batch_size = 2
    frames = 2
    x = torch.ones((batch_size, frames, 1, stimulus.n_input_elements))
    stimulus.add_input(x)
    assert stimulus.buffer.shape == (batch_size, frames, stimulus.n_nodes)
    assert stimulus.buffer.sum() == batch_size * frames * np.prod(
        stimulus.input_index.shape
    )

    # single batch, temporal start
    batch_size = 1
    frames = 20
    start = 10
    x = torch.ones((batch_size, frames - start, 1, stimulus.n_input_elements))
    stimulus.add_input(x, start=start, n_frames_buffer=frames)
    assert stimulus.buffer.shape == (batch_size, frames, stimulus.n_nodes)
    assert stimulus.buffer.sum() == batch_size * (frames - start) * np.prod(
        stimulus.input_index.shape
    )

    # single batch, temporal start and stop
    batch_size = 1
    frames = 20
    start = 5
    stop = 15
    x = torch.ones((batch_size, stop - start, 1, stimulus.n_input_elements))
    stimulus.add_input(x, start=start, stop=stop, n_frames_buffer=frames)
    assert stimulus.buffer.shape == (batch_size, frames, stimulus.n_nodes)
    assert stimulus.buffer.sum() == batch_size * (stop - start) * np.prod(
        stimulus.input_index.shape
    )

    # cumulate
    stimulus.add_input(x, start=start, stop=stop, n_frames_buffer=frames, cumulate=True)
    assert stimulus.buffer.sum() == 2 * batch_size * (stop - start) * np.prod(
        stimulus.input_index.shape
    )

    # don't cumulate
    stimulus.add_input(x, start=start, stop=stop, n_frames_buffer=frames, cumulate=False)
    assert stimulus.buffer.sum() == batch_size * (stop - start) * np.prod(
        stimulus.input_index.shape
    )

    # fixed_frames - actual frames mismatch raises error
    batch_size = 1
    frames = 20
    x = torch.ones((batch_size, frames + 1, 1, stimulus.n_input_elements))
    with pytest.raises(RuntimeError):
        stimulus.add_input(x, n_frames_buffer=frames)

    # without specifying fixed_frames, resizes buffer
    stimulus.add_input(x)
    assert stimulus.buffer.shape == (batch_size, frames + 1, stimulus.n_nodes)
    assert stimulus.buffer.sum() == batch_size * (frames + 1) * np.prod(
        stimulus.input_index.shape
    )
    # since cumulate is False by default and the buffer has nonzero elements
    # the buffer should be initialized to zero
    assert stimulus.nonzero
    stimulus.add_input(x)
    assert stimulus.buffer.sum() == batch_size * (frames + 1) * np.prod(
        stimulus.input_index.shape
    )

    # setting the _nonzero attribute to False should allow the buffer to
    # accumulate even though cumulate is False by default
    stimulus._nonzero = False
    stimulus.add_input(x)
    assert stimulus.buffer.sum() == 2 * batch_size * (frames + 1) * np.prod(
        stimulus.input_index.shape
    )


def test_add_pre_stim(stimulus: Stimulus):
    # adding single value
    stimulus.add_pre_stim(1.0)
    assert stimulus.buffer.shape == (1, 1, stimulus.n_nodes)
    assert stimulus.buffer.sum() == np.prod(stimulus.input_index.shape)

    # adding tensor but not matching implicit buffer
    x = torch.ones((10))
    with pytest.raises(RuntimeError):
        stimulus.add_pre_stim(x)

    # adding tensor and not matching explicit buffer
    stimulus.zero(1, 20)
    with pytest.raises(RuntimeError):
        stimulus.add_pre_stim(x)

    # adding tensor and matching explicit buffer
    stimulus.zero(1, 10)
    stimulus.add_pre_stim(x)
    assert stimulus.buffer.shape == (1, 10, stimulus.n_nodes)
    assert stimulus.buffer.sum() == 10 * np.prod(stimulus.input_index.shape)

    # adding tensor and matching explicit buffer using start, stop and n_frames_buffer
    stimulus.add_pre_stim(x, start=0, stop=10, n_frames_buffer=20)
    assert stimulus.buffer.shape == (1, 20, stimulus.n_nodes)
    assert stimulus.buffer.sum() == 10 * np.prod(stimulus.input_index.shape)



"""Ensemble of trained networks."""

from __future__ import annotations

import logging
import os
import pickle
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import dataclass
from functools import wraps
from pathlib import Path
from typing import (
    Callable,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

import numpy as np
import torch
import xarray as xr
from cachetools import FIFOCache
from matplotlib import colormaps as cm
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Colormap, Normalize
from numpy.typing import NDArray
from torch import nn
from tqdm.auto import tqdm

import flyvis
from flyvis.analysis import stimulus_responses, stimulus_responses_currents
from flyvis.analysis.clustering import (
    GaussianMixtureClustering,
    compute_umap_and_clustering,
    get_cluster_to_indices,
)
from flyvis.analysis.visualization import plots
from flyvis.connectome import get_avgfilt_connectome
from flyvis.utils.cache_utils import context_aware_cache
from flyvis.utils.chkpt_utils import (
    best_checkpoint_default_fn,
    recover_network,
    resolve_checkpoints,
)
from flyvis.utils.class_utils import forward_subclass
from flyvis.utils.logging_utils import all_logging_disabled
from flyvis.utils.nn_utils import simulation

from .directories import EnsembleDir, NetworkDir
from .initialization import Parameter
from .network import Network
from .network_view import NetworkView

logging = logging.getLogger(__name__)

__all__ = ["Ensemble"]


class Ensemble(dict):
    """Dictionary to a collection of trained networks.

    Args:
        path: Path to ensemble directory or list of paths to model directories.
            Can be a single string, then assumes the path is the root directory
            as configured by datamate.
        network_class: Class to use for initializing networks.
        root_dir: Root directory for model paths.
        connectome_getter: Function to get the connectome.
        checkpoint_mapper: Function to map checkpoints.
        best_checkpoint_fn: Function to determine best checkpoint.
        best_checkpoint_fn_kwargs: Kwargs for best_checkpoint_fn.
        recover_fn: Function to recover network.
        try_sort: Whether to try to sort the ensemble by validation error.

    Attributes:
        names (List[str]): List of model names.
        name (str): Ensemble name.
        path (Path): Path to ensemble directory.
        model_paths (List[Path]): List of paths to model directories.
        dir (EnsembleDir): Directory object for ensemble directory.

    Note:
        The ensemble is a dynamic dictionary, so you can access the networks
        in the ensemble by name or index. For example, to access the first network
        simply do:
        ```python
        ensemble[0]
        ```
        or
        ```python
        ensemble['flow/000/0000']
        ```
        Slice to create a subset of the ensemble:
        ```python
        ensemble[0:2]
        ```
    """

    def __init__(
        self,
        path: Union[str, Path, Iterable, "EnsembleDir"],
        network_class: nn.Module = Network,
        root_dir: Path = flyvis.results_dir,
        connectome_getter: Callable = get_avgfilt_connectome,
        checkpoint_mapper: Callable = resolve_checkpoints,
        best_checkpoint_fn: Callable = best_checkpoint_default_fn,
        best_checkpoint_fn_kwargs: dict = {
            "validation_subdir": "validation",
            "loss_file_name": "epe",
        },
        recover_fn: Callable = recover_network,
        try_sort: bool = False,
    ):
        if isinstance(path, EnsembleDir):
            path = path.path
            self.model_paths, self.path = model_paths_from_parent(path)
            self.dir = path
        elif isinstance(path, Path):
            self.model_paths, self.path = model_paths_from_parent(path)
            self.dir = EnsembleDir(self.path)
        elif isinstance(path, str):
            self.dir = EnsembleDir(path)
            self.model_paths, self.path = model_paths_from_parent(self.dir.path)
        elif isinstance(path, Iterable):
            self.model_paths, self.path = model_paths_from_names_or_paths(path, root_dir)
            self.dir = EnsembleDir(self.path)
        else:
            raise TypeError(
                f"Unsupported path type: {type(path)}. "
                "Expected EnsembleDir, str, PathLike, or Iterable."
            )

        self.names, self.name = model_path_names(self.model_paths)
        self.in_context = False

        self._names = []
        self.model_index = []
        # Initialize pointers to model directories.
        for i, name in tqdm(
            enumerate(self.names), desc="Loading ensemble", total=len(self.names)
        ):
            try:
                with all_logging_disabled():
                    self[name] = NetworkView(
                        NetworkDir(self.model_paths[i]),
                        network_class=network_class,
                        root_dir=root_dir,
                        connectome_getter=connectome_getter,
                        checkpoint_mapper=checkpoint_mapper,
                        best_checkpoint_fn=best_checkpoint_fn,
                        best_checkpoint_fn_kwargs=best_checkpoint_fn_kwargs,
                        recover_fn=recover_fn,
                    )
                    self._names.append(name)
            except AttributeError as e:
                logging.warning(f"Failed to load {name}: {e}")
        self._broken = list(set(self.names) - set(self._names))
        self.names = self._names
        self.model_index = np.arange(len(self.names))
        logging.info(f"Loaded {len(self)} networks.")

        if try_sort:
            self.sort()

        self._init_args = (
            path,
            network_class,
            root_dir,
            connectome_getter,
            checkpoint_mapper,
            best_checkpoint_fn,
            best_checkpoint_fn_kwargs,
            recover_fn,
            try_sort,
        )
        self.connectome = self[next(iter(self))].connectome
        self.cache = FIFOCache(maxsize=3)

    def __getitem__(
        self, key: Union[str, int, slice, NDArray, list]
    ) -> Union[NetworkView, "Ensemble"]:
        """Get item from the ensemble.

        Args:
            key: Key to access the ensemble.
                Can be a string, int, slice, NDArray, or list.

        Returns:
            NetworkView or Ensemble: The requested item or subset of the ensemble.

        Raises:
            ValueError: If the key is invalid.
        """
        if isinstance(key, (int, np.integer)):
            return dict.__getitem__(self, self.names[key])
        elif isinstance(key, slice):
            return self.__class__(self.names[key])
        elif isinstance(key, (np.ndarray, list)):
            return self.__class__(np.array(self.names)[key])
        elif key in self.names:
            return dict.__getitem__(self, key)
        else:
            raise ValueError(f"{key}")

    def __repr__(self) -> str:
        """Return a string representation of the Ensemble."""
        return f"{self.__class__.__name__}({self.path})"

    def __dir__(self) -> List[str]:
        """Return a list of attributes for the Ensemble."""
        return list({*dict.__dir__(self), *dict.__iter__(self)})

    def __len__(self) -> int:
        """Return the number of networks in the Ensemble."""
        return len(self.names)

    def __iter__(self) -> Iterator[str]:
        """Iterate over the names of the networks in the Ensemble."""
        yield from self.names

    def items(self) -> Iterator[Tuple[str, NetworkView]]:
        """Return an iterator over the (name, NetworkView) pairs in the Ensemble."""
        return iter((k, self[k]) for k in self)

    def keys(self) -> List[str]:
        """Return a list of network names in the Ensemble."""
        return list(self)

    def values(self) -> List[NetworkView]:
        """Return a list of NetworkViews in the Ensemble."""
        return [self[k] for k in self]

    def _clear_cache(self) -> None:
        """Clear the cache of the Ensemble."""
        self.cache = {}

    def _clear_memory(self) -> None:
        """Clear the memory of all NetworkViews in the Ensemble."""
        for nv in self.values():
            nv._clear_memory()

    def check_configs_match(self) -> bool:
        """Check if the configurations of the networks in the ensemble match.

        Returns:
            bool: True if all configurations match, False otherwise.
        """
        config0 = self[0].dir.config
        for i in range(1, len(self)):
            diff = config0.diff(self[i].dir.config, name1="first", name2="second").first
            if diff and not (len(diff) == 1 and "network_name" in diff[0]):
                logging.warning(
                    "%(first)s differs from %(second)s. Diff is %(diff)s.",
                    {"first": self[0].name, "second": self[i].name, "diff": diff},
                )
                return False
        return True

    def yield_networks(self) -> Generator[Network, None, None]:
        """Yield initialized networks from the ensemble.

        Yields:
            Network: Initialized network from the ensemble.
        """
        network = self[0].init_network()
        yield network
        for network_view in self.values()[1:]:
            yield network_view.init_network(network=network)

    def yield_decoders(self) -> Generator[nn.Module, None, None]:
        """Yield initialized decoders from the ensemble.

        Yields:
            nn.Module: Initialized decoder from the ensemble.
        """
        assert self.check_configs_match(), "configurations do not match"
        decoder = self[0].init_decoder()
        for network_view in self.values():
            yield network_view.init_decoder(decoder=decoder)

    def simulate(
        self, movie_input: torch.Tensor, dt: float, fade_in: bool = True
    ) -> Generator[np.ndarray, None, None]:
        """Simulate the ensemble activity from movie input.

        Args:
            movie_input: Tensor with shape (batch_size, n_frames, 1, hexals).
            dt: Integration time constant. Warns if dt > 1/50.
            fade_in: Whether to use `network.fade_in_state` to compute the initial
                state. Defaults to True. If False, uses the
                `network.steady_state` after 1s of grey input.

        Yields:
            np.ndarray: Response of each individual network.

        Note:
            Simulates across batch_size in parallel, which can easily lead to OOM for
            large batch sizes.
        """
        for network in tqdm(
            self.yield_networks(),
            desc="Simulating network",
            total=len(self.names),
        ):
            yield (
                network.simulate(
                    movie_input,
                    dt,
                    initial_state=(
                        network.fade_in_state(1.0, dt, movie_input[:, 0])
                        if fade_in
                        else "auto"
                    ),
                )
                .cpu()
                .numpy()
            )

    def simulate_from_dataset(
        self,
        dataset,
        dt: float,
        indices: Iterable[int] = None,
        t_pre: float = 1.0,
        t_fade_in: float = 0.0,
        default_stim_key: str = "lum",
        batch_size: int = 1,
        central_cell_only: bool = True,
    ) -> Generator[np.ndarray, None, None]:
        """Simulate the ensemble activity from a dataset.

        Args:
            dataset: Dataset to simulate from.
            dt: Integration time constant.
            indices: Indices of stimuli to simulate. Defaults to None (all stimuli).
            t_pre: Time before stimulus onset. Defaults to 1.0.
            t_fade_in: Fade-in time. Defaults to 0.0.
            default_stim_key: Default stimulus key. Defaults to "lum".
            batch_size: Batch size for simulation. Defaults to 1.
            central_cell_only: Whether to return only central cells. Defaults to True.

        Yields:
            np.ndarray: Simulated responses for each network.
        """
        if central_cell_only:
            central_cells_index = self[0].connectome.central_cells_index[:]

        progress_bar = tqdm(desc="Simulating network", total=len(self.names))

        for network in self.yield_networks():

            def handle_network(network: Network):
                for _, resp in network.stimulus_response(
                    dataset,
                    dt=dt,
                    indices=indices,
                    t_pre=t_pre,
                    t_fade_in=t_fade_in,
                    default_stim_key=default_stim_key,
                    batch_size=batch_size,
                ):
                    if central_cell_only:
                        yield resp[:, :, central_cells_index]
                    else:
                        yield resp

            r = np.stack(list(handle_network(network)))
            yield r.reshape(-1, r.shape[-2], r.shape[-1])

            progress_bar.update(1)
        progress_bar.close()

    def decode(
        self, movie_input: torch.Tensor, dt: float
    ) -> Generator[np.ndarray, None, None]:
        """Decode the ensemble responses with the ensemble decoders.

        Args:
            movie_input: Input movie tensor.
            dt: Integration time constant.

        Yields:
            np.ndarray: Decoded responses for each network.
        """
        responses = torch.tensor(list(self.simulate(movie_input, dt)))
        for i, decoder in enumerate(self.yield_decoders()):
            with simulation(decoder):
                yield decoder(responses[i]).cpu().numpy()

    def validation_file(
        self,
        validation_subdir: Optional[str] = None,
        loss_file_name: Optional[str] = None,
    ) -> Tuple[str, str]:
        """Return the validation file for each network in the ensemble.

        Args:
            validation_subdir: Subdirectory for validation files. Defaults to None.
            loss_file_name: Name of the loss file. Defaults to None.

        Returns:
            Tuple[str, str]: Validation subdirectory and loss file name.
        """
        network_view0 = self[0]
        if validation_subdir is None:
            validation_subdir = network_view0.best_checkpoint_fn_kwargs.get(
                "validation_subdir"
            )
        if loss_file_name is None:
            loss_file_name = network_view0.best_checkpoint_fn_kwargs.get("loss_file_name")

        return validation_subdir, loss_file_name

    def sort(
        self,
        validation_subdir: Optional[str] = None,
        loss_file_name: Optional[str] = None,
    ) -> None:
        """Sort the ensemble by validation loss.

        Args:
            validation_subdir: Subdirectory for validation files. Defaults to None.
            loss_file_name: Name of the loss file. Defaults to None.
        """
        try:
            self.names = sorted(
                self.keys(),
                key=lambda key: dict(
                    zip(
                        self.keys(),
                        self.min_validation_losses(validation_subdir, loss_file_name),
                    )
                )[key],
                reverse=False,
            )
        except Exception as e:
            logging.info(f"sorting failed: {e}")

    def argsort(
        self,
        validation_subdir: Optional[str] = None,
        loss_file_name: Optional[str] = None,
    ) -> np.ndarray:
        """Return the indices that would sort the ensemble by validation loss.

        Args:
            validation_subdir: Subdirectory for validation files. Defaults to None.
            loss_file_name: Name of the loss file. Defaults to None.

        Returns:
            np.ndarray: Indices that would sort the ensemble.
        """
        return np.argsort(
            self.min_validation_losses(
                *self.validation_file(validation_subdir, loss_file_name)
            )
        )

    def zorder(
        self,
        validation_subdir: Optional[str] = None,
        loss_file_name: Optional[str] = None,
    ) -> np.ndarray:
        """Return the z-order of the ensemble based on validation loss.

        Args:
            validation_subdir: Subdirectory for validation files. Defaults to None.
            loss_file_name: Name of the loss file. Defaults to None.

        Returns:
            np.ndarray: Z-order of the ensemble.
        """
        return len(self) - self.argsort(validation_subdir, loss_file_name).argsort()

    @context_aware_cache(context=lambda self: (self.names))
    def validation_losses(
        self, subdir: Optional[str] = None, file: Optional[str] = None
    ) -> np.ndarray:
        """Return a list of validation losses for each network in the ensemble.

        Args:
            subdir: Subdirectory for validation files. Defaults to None.
            file: Name of the loss file. Defaults to None.

        Returns:
            np.ndarray: Validation losses for each network.
        """
        subdir, file = self.validation_file(subdir, file)
        losses = np.array([nv.dir[subdir][file][()] for nv in self.values()])
        return losses

    def min_validation_losses(
        self, subdir: Optional[str] = None, file: Optional[str] = None
    ) -> np.ndarray:
        """Return the minimum validation loss of the ensemble.

        Args:
            subdir: Subdirectory for validation files. Defaults to None.
            file: Name of the loss file. Defaults to None.

        Returns:
            np.ndarray: Minimum validation losses for each network.
        """
        losses = self.validation_losses(subdir, file)
        if losses.ndim == 1:
            return losses
        return np.min(losses, axis=1)

    @contextmanager
    def rank_by_validation_error(self, reverse: bool = False):
        """Temporarily sort the ensemble based on validation error.

        Args:
            reverse: Whether to sort in descending order. Defaults to False.

        Yields:
            None
        """
        _names = deepcopy(self.names)

        try:
            self.names = sorted(
                self.keys(),
                key=lambda key: dict(zip(self.keys(), self.min_validation_losses()))[key],
                reverse=reverse,
            )
        except Exception as e:
            logging.info(f"sorting failed: {e}")
        try:
            yield
        finally:
            self.names = list(_names)

    @contextmanager
    def ratio(self, best: Optional[float] = None, worst: Optional[float] = None):
        """Temporarily filter the ensemble by a ratio of best or worst performing models.

        Args:
            best: Ratio of best performing models to keep. Defaults to None.
            worst: Ratio of worst performing models to keep. Defaults to None.

        Yields:
            None

        Raises:
            ValueError: If best and worst sum to more than 1.
        """
        # no-op
        if best is None and worst is None:
            yield
            return

        _names = tuple(self.names)
        _model_index = tuple(self.model_index)

        with self.rank_by_validation_error():
            if best is not None and worst is not None and best + worst > 1:
                raise ValueError("best and worst must add up to 1")

            if best is not None or worst is not None:
                _context_best_names, _context_worst_names = [], []
                if best is not None:
                    _context_best_names = list(self.names[: int(best * len(self))])
                    self._best_ratio = best
                else:
                    self._best_ratio = 0
                if worst is not None:
                    _context_worst_names = list(self.names[-int(worst * len(self)) :])
                    self._worst_ratio = worst
                else:
                    self._worst_ratio = 0

                in_context_names = [*_context_best_names, *_context_worst_names]

                if in_context_names:  # to prevent an empty index
                    self.model_index = np.array([
                        i
                        for i, name in zip(_model_index, _names)
                        if name in in_context_names
                    ])
                self.names = in_context_names
                self.in_context = True
            try:
                yield
            finally:
                self.names = list(_names)
                self.model_index = _model_index
                self._best_ratio = 0.5
                self._worst_ratio = 0.5
                self.in_context = False

    @contextmanager
    def select_items(self, indices: List[int]):
        """Temporarily filter the ensemble by a list of indices.

        Args:
            indices: List of indices to select.

        Yields:
            None

        Raises:
            ValueError: If indices are invalid.
        """
        # no-op
        try:
            if indices is None:
                yield
                return
            _names = tuple(self.names)
            _model_index = tuple(self.model_index)
            self._names = _names

            if isinstance(indices, (int, np.integer, slice)):
                in_context_names = self.names[indices]
            elif isinstance(indices, (list, np.ndarray)):
                if np.array(indices).dtype == np.array(self.names).dtype:
                    in_context_names = indices
                elif np.array(indices).dtype == np.int_:
                    in_context_names = np.array(self.names)[indices]
                else:
                    raise ValueError(f"{indices}")
            else:
                raise ValueError(f"{indices}")
            self.model_index = np.array([
                i for i, name in zip(_model_index, _names) if name in in_context_names
            ])
            self.names = in_context_names
            self.in_context = True
            yield
        finally:
            self.names = list(_names)
            self.model_index = list(_model_index)
            self.in_context = False

    def task_error(
        self,
        cmap: Union[str, Colormap] = "Blues_r",
        truncate: Optional[Dict[str, Union[float, int]]] = None,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
    ) -> "TaskError":
        """Return a TaskError object for the ensemble.

        Args:
            cmap: Colormap to use. Defaults to "Blues_r".
            truncate: Dictionary to truncate the colormap. Defaults to None.
            vmin: Minimum value for normalization. Defaults to None.
            vmax: Maximum value for normalization. Defaults to None.

        Returns:
            TaskError: Object containing validation losses, colors, colormap, norm,
                and scalar mapper.
        """
        error = self.min_validation_losses()

        if truncate is None:
            # truncate because the maxval would be white with the default colormap
            # which would be invisible on a white background
            truncate = {"minval": 0.0, "maxval": 0.9, "n": 256}
        cmap = cm.get_cmap(cmap) if isinstance(cmap, str) else cmap
        cmap = plots.plt_utils.truncate_colormap(cmap, **truncate)
        sm, norm = plots.plt_utils.get_scalarmapper(
            cmap=cmap,
            vmin=vmin or np.min(error),
            vmax=vmax or np.max(error),
        )
        colors = sm.to_rgba(np.array(error))

        return TaskError(error, colors, cmap, norm, sm)

    def parameters(self) -> Dict[str, np.ndarray]:
        """Return the parameters of the ensemble.

        Returns:
            Dict[str, np.ndarray]: Dictionary of parameter arrays.
        """
        network_params = {}
        for network_view in self.values():
            chkpt_params = torch.load(network_view.network('best').checkpoint)
            for key, val in chkpt_params["network"].items():
                if key not in network_params:
                    network_params[key] = []
                network_params[key].append(val.cpu().numpy())
        for key, val in network_params.items():
            network_params[key] = np.array(val)
        return network_params

    def parameter_keys(self) -> Dict[str, List[str]]:
        """Return the keys of the parameters of the ensemble.

        Returns:
            Dict[str, List[str]]: Dictionary of parameter keys.
        """
        self.check_configs_match()
        network_view = self[0]
        config = network_view.dir.config.network

        parameter_keys = {}
        for param_name, param_config in config.node_config.items():
            param = forward_subclass(
                Parameter,
                config={
                    "type": param_config.type,
                    "param_config": param_config,
                    "connectome": network_view.connectome,
                },
            )
            parameter_keys[f"nodes_{param_name}"] = param.keys
        for param_name, param_config in config.edge_config.items():
            param = forward_subclass(
                Parameter,
                config={
                    "type": param_config.type,
                    "param_config": param_config,
                    "connectome": network_view.connectome,
                },
            )
            parameter_keys[f"edges_{param_name}"] = param.keys
        return parameter_keys

    @wraps(stimulus_responses.flash_responses)
    @context_aware_cache(context=lambda self: (self.names))
    def flash_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate flash responses."""
        return stimulus_responses.flash_responses(self, *args, **kwargs)

    @wraps(stimulus_responses.moving_edge_responses)
    @context_aware_cache(context=lambda self: (self.names))
    def moving_edge_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate moving edge responses."""
        return stimulus_responses.moving_edge_responses(self, *args, **kwargs)

    @wraps(stimulus_responses.moving_bar_responses)
    @context_aware_cache(context=lambda self: (self.names))
    def moving_bar_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate moving bar responses."""
        return stimulus_responses.moving_bar_responses(self, *args, **kwargs)

    @wraps(stimulus_responses.naturalistic_stimuli_responses)
    @context_aware_cache(context=lambda self: (self.names))
    def naturalistic_stimuli_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate naturalistic stimuli responses."""
        return stimulus_responses.naturalistic_stimuli_responses(self, *args, **kwargs)

    @wraps(stimulus_responses.central_impulses_responses)
    @context_aware_cache(context=lambda self: (self.names))
    def central_impulses_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate central ommatidium impulses responses."""
        return stimulus_responses.central_impulses_responses(self, *args, **kwargs)

    @wraps(stimulus_responses.spatial_impulses_responses)
    @context_aware_cache(context=lambda self: (self.names))
    def spatial_impulses_responses(self, *args, **kwargs) -> xr.Dataset:
        """Generate spatial ommatidium impulses responses."""
        return stimulus_responses.spatial_impulses_responses(self, *args, **kwargs)

    @wraps(stimulus_responses_currents.moving_edge_currents)
    @context_aware_cache(context=lambda self: (self.names))
    def moving_edge_currents(
        self, *args, **kwargs
    ) -> List[stimulus_responses_currents.ExperimentData]:
        """Generate moving edge currents."""
        return stimulus_responses_currents.moving_edge_currents(self, *args, **kwargs)

    @context_aware_cache
    def clustering(self, cell_type) -> GaussianMixtureClustering:
        """Return the clustering of the ensemble for a given cell type.

        Args:
            cell_type: The cell type to cluster.

        Returns:
            GaussianMixtureClustering: Clustering object for the given cell type.

        Raises:
            ValueError: If clustering is not available in context.
        """
        if self.in_context:
            raise ValueError("clustering is not available in context")

        if (
            not self.dir.umap_and_clustering
            or not self.dir.umap_and_clustering[cell_type]
        ):
            return compute_umap_and_clustering(self, cell_type)

        path = self.dir.umap_and_clustering[f"{cell_type}.pickle"]
        with open(path, "rb") as file:
            clustering = pickle.load(file)

        return clustering

    def cluster_indices(self, cell_type: str) -> Dict[int, NDArray[int]]:
        """Clusters from responses to naturalistic stimuli of the given cell type.

        Args:
            cell_type: The cell type to return the clusters for.

        Returns:
            Dict[int, NDArray[int]]: Keys are the cluster ids and the values are the
            model indices in the ensemble.

        Example:
            ```python
            ensemble = Ensemble("path/to/ensemble")
            cluster_indices = ensemble.cluster_indices("T4a")
            first_cluster = ensemble[cluster_indices[0]]
            ```

        Raises:
            ValueError: If stored clustering does not match ensemble.
        """
        clustering = self.clustering(cell_type)
        cluster_indices = get_cluster_to_indices(
            clustering.embedding.mask,
            clustering.labels,
            task_error=self.task_error(),
        )

        _models = sorted(np.concatenate(list(cluster_indices.values())))
        if len(_models) != clustering.embedding.mask.sum() or len(_models) > len(self):
            raise ValueError("stored clustering does not match ensemble")

        return cluster_indices

    def responses_norm(self, rectified: bool = False) -> np.ndarray:
        """Compute the norm of responses to naturalistic stimuli.

        Args:
            rectified: Whether to rectify responses before computing norm.

        Returns:
            np.ndarray: Norm of responses for each network.
        """
        response_set = self.naturalistic_stimuli_responses()
        responses = response_set['responses'].values

        def compute_norm(X, rectified=True):
            """Computes a normalization constant for stimulus
                responses per cell hypothesis, i.e. cell_type independent values.

            Args:
                X: (n_stimuli, n_frames, n_cell_types)
            """
            if rectified:
                X = np.maximum(X, 0)
            n_models, n_samples, n_frames, n_cell_types = X.shape

            # replace NaNs with 0
            X[np.isnan(X)] = 0

            return (
                1
                / np.sqrt(n_samples * n_frames)
                * np.linalg.norm(
                    X,
                    axis=(1, 2),
                    keepdims=True,
                )
            )

        return np.take(
            compute_norm(responses, rectified=rectified), self.model_index, axis=0
        )


def model_path_names(model_paths: List[Path]) -> Tuple[List[str], str]:
    """Return a list of model names and an ensemble name from a list of model paths.

    Args:
        model_paths: List of paths to model directories.

    Returns:
        Tuple[List[str], str]: List of model names and ensemble name.
    """
    model_names = [os.path.sep.join(path.parts[-3:]) for path in model_paths]
    ensemble_name = ", ".join(
        np.unique([
            os.path.sep.join(n.split(os.path.sep)[:2]) for n in model_names
        ]).tolist()
    )
    return model_names, ensemble_name


def model_paths_from_parent(path: Path) -> Tuple[List[Path], Path]:
    """Return a list of model paths from a parent path.

    Args:
        path: Parent path to search for model directories.

    Returns:
        Tuple[List[Path], Path]: List of model paths and parent path.
    """
    model_paths = sorted(
        filter(
            lambda p: p.name.isnumeric() and p.is_dir(),
            path.iterdir(),
        )
    )
    return model_paths, path


def model_paths_from_names_or_paths(
    paths: List[Union[str, Path]], root_dir: Path
) -> Tuple[List[Path], Path]:
    """Return model paths and ensemble path from model names or paths.

    Args:
        paths: List of model names or paths.
        root_dir: Root directory for model paths.

    Returns:
        Tuple[List[Path], Path]: List of model paths and ensemble path.

    Raises:
        ValueError: If an invalid path type is provided.
        NotImplementedError: If multiple ensemble paths are found.
    """
    model_paths = []
    _ensemble_paths = []
    for path in paths:
        if isinstance(path, str):
            path_obj = Path(path)
            path_parts = path_obj.parts
            # assuming task/ensemble_id/model_id
            if len(path_parts) == 3:
                model_paths.append(root_dir / path_obj)
            # assuming task/ensemble_id
            elif len(path_parts) == 2:
                model_paths.extend(model_paths_from_parent(path_obj)[0])
        elif isinstance(path, Path):
            model_paths.append(path)
        else:
            raise ValueError(f"Invalid path type: {path}")
        _ensemble_paths.append(model_paths[-1].parent)
    # Ensure all ensemble paths are the same
    ensemble_paths_set = set(_ensemble_paths)
    if len(ensemble_paths_set) != 1:
        raise NotImplementedError("Multiple ensemble paths found")
    ensemble_path = _ensemble_paths[0]
    return model_paths, ensemble_path


@dataclass
class TaskError:
    """A dataclass that contains the validation losses, colors, colormap, norm,
    and scalar mapper.

    Attributes:
        values (NDArray): Validation loss values.
        colors (NDArray): Colors corresponding to validation losses.
        cmap (Colormap): Colormap used for visualization.
        norm (Normalize): Normalization object for the colormap.
        scalarmappable (ScalarMappable): ScalarMappable object for the colormap.
    """

    values: NDArray
    colors: NDArray
    cmap: Colormap
    norm: Normalize
    scalarmappable: ScalarMappable



from datamate import Namespace, set_root_context

from flyvis.task.tasks import Task


def test_task(mock_sintel_data, connectome, tmp_path_factory):
    task_config = Namespace(
        type="Task",
        dataset=Namespace(
            type="MultiTaskSintel",
            tasks=["flow"],
            boxfilter=dict(extent=15, kernel_size=13),
            vertical_splits=3,
            n_frames=19,
            center_crop_fraction=0.7,
            dt=1 / 50,
            augment=True,
            random_temporal_crop=True,
            all_frames=False,
            resampling=True,
            interpolate=True,
            p_flip=0.5,
            p_rot=5 / 6,
            contrast_std=0.2,
            brightness_std=0.1,
            gaussian_white_noise=0.08,
            gamma_std=None,
            _init_cache=True,
            unittest=True,
            flip_axes=[0, 1, 2, 3],
            sintel_path=str(mock_sintel_data),
        ),
        decoder=Namespace(
            flow=Namespace(
                type="DecoderGAVP",
                shape=[8, 2],
                kernel_size=5,
                const_weight=0.001,
                n_out_features=None,
                p_dropout=0.5,
            )
        ),
        loss=Namespace(flow="l2norm"),
        batch_size=4,
        num_workers=0,
        n_iters=250000,
        train_ratio=None,
        seed=1,
        transforms=None,
        train_seq_ind=None,
        val_seq_ind=None,
    )
    with set_root_context(tmp_path_factory.mktemp("tmp")):
        task = Task(
            task_config.dataset,
            task_config.decoder,
            task_config.loss,
            batch_size=4,
            n_iters=250_000,
            n_folds=4,
            fold=1,
            seed=0,
        )
        assert task is not None
        assert task.dataset is not None

        decoder = task.init_decoder(connectome)
        assert decoder is not None



import inspect
import logging

import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

import flyvis
from flyvis.datasets import MultiTaskDataset
from flyvis.network import Network, NetworkView
from flyvis.task.objectives import epe, l2norm
from flyvis.utils.class_utils import forward_subclass

logging = logging.getLogger(__name__)

__all__ = ["validate", "validate_all_checkpoints"]


@torch.no_grad()
def validate(
    network: Network,
    decoder,
    dataloader,
    loss_fns,
    dt,
    t_pre=0.0,
    loss_kwargs={},
):
    """Tests the network and decoder on the dataloader with the given loss functions."""
    network.eval()

    for _decoder in decoder.values():
        _decoder.eval()

    dataset = dataloader.dataset
    steady_state = network.steady_state(
        t_pre=t_pre,
        dt=dt,
        batch_size=dataloader.batch_size,
        value=0.5,
        state=None,
        grad=False,
    )
    losses = {task: [] for task in dataset.tasks}  # type: Dict[str, List]
    stimulus = network.stimulus

    with dataset.augmentation(False):
        for _, data in enumerate(dataloader):
            # Resets the stimulus buffer (#frames, #samples, #neurons).
            # The number of frames and samples can change, but the number of nodes is
            # constant.
            n_samples, n_frames, _, _ = data["lum"].shape
            stimulus.zero(n_samples, n_frames)

            # Add batch of hex-videos (#frames, #samples, #hexals) as photorecptor
            # stimuli.
            stimulus.add_input(data["lum"])

            # Run stimulus through network.
            activity = network(stimulus(), dt, state=steady_state)

            # Decode activity and evaluate loss.
            for task in dataset.tasks:
                y = data[task]
                y_est = decoder[task](activity)
                losses[task].append([
                    fn(y_est, y, **loss_kwargs).detach().cpu().item() for fn in loss_fns
                ])

    summed_loss = 0
    task_loss = {}
    # Record loss per task (+rec).
    for task in losses:
        # (#samples, #loss_functions)
        loss = np.array(losses[task])
        sample_average_loss = np.mean(loss, axis=0)
        if f"loss_{task}" not in task_loss:
            task_loss[f"loss_{task}"] = []
        task_loss[f"loss_{task}"].append(sample_average_loss)
        summed_loss += sample_average_loss
    # Record average loss.
    val_loss = summed_loss / len(losses)

    network.train()
    for _decoder in decoder.values():
        _decoder.train()

    return val_loss, task_loss


def validate_all_checkpoints(
    network_view: NetworkView,
    loss_fns=None,
    dt=1 / 50,
    t_pre=0.5,
    validation_subdir="validation",
):
    dataset = forward_subclass(MultiTaskDataset, network_view.dir.config.task.dataset)
    loss_fns = get_loss_fns(loss_fns)
    network_view.init_network()
    network_view.init_decoder()
    _, val_sequences = dataset.original_train_and_validation_indices()

    dataloader = DataLoader(
        dataset,
        batch_size=1,
        num_workers=0,
        sampler=flyvis.utils.dataset_utils.IndexSampler(val_sequences),
        drop_last=False,
    )

    dataloader.dataset.dt = dt

    loss = []
    progress = tqdm(total=len(network_view.dir.chkpt_index))
    for chkpt in network_view.checkpoints.indices:
        network = network_view.network(checkpoint=chkpt)
        decoder = network_view.init_decoder(
            checkpoint=chkpt, decoder=network_view.decoder
        )
        loss.append(
            validate(
                network=network.network,
                decoder=decoder,
                dataloader=dataloader,
                loss_fns=loss_fns,
                dt=dt,
                t_pre=t_pre,
            )[0]
        )
        progress.update(1)
    progress.close()

    loss = np.array(loss)

    for i, fn in enumerate(loss_fns):
        network_view.dir[validation_subdir][fn.__name__] = loss[:, i]

    network_view.dir[validation_subdir].config = dict(
        dt=dt,
        t_pre=t_pre,
        loss_fns=[fn.__name__ for fn in loss_fns],
        validation_subdir=validation_subdir,
        validation_function=inspect.currentframe().f_code.co_name,
    )

    return loss


def get_loss_fns(loss_fns):
    if loss_fns:
        return loss_fns
    return [l2norm, epe]



import math

import numpy as np
import torch

from flyvis.datasets.augmentation.augmentation import Augmentation
from flyvis.datasets.augmentation.hex import (
    ContrastBrightness,
    HexFlip,
    HexRotate,
    PixelNoise,
)
from flyvis.datasets.augmentation.temporal import (
    CropFrames,
    # InterpolateFrames,
    Interpolate,
)


def test_base_class():
    class Test(Augmentation):
        def transform(self, sequence: torch.Tensor):
            return sequence * 2

    class Test2(Augmentation):
        def transform(self, sequence: torch.Tensor):
            return sequence * 3

    class Test3(Test2):
        def transform(self, sequence: torch.Tensor):
            return sequence * 4

    test0 = Augmentation()
    test1 = Test()
    test2 = Test2()
    test3 = Test3()
    sequence = torch.tensor([0, 1])

    assert test0(sequence).tolist() == [0, 1]
    test0.augment = False
    assert test0(sequence).tolist() == [0, 1]

    assert test1(sequence).tolist() == [0, 2]
    test1.augment = False
    assert test1(sequence).tolist() == [0, 1]

    assert test2(sequence).tolist() == [0, 3]
    test2.augment = False
    assert test2(sequence).tolist() == [0, 1]

    assert test3(sequence).tolist() == [0, 4]
    test3.augment = False
    assert test3(sequence).tolist() == [0, 1]


# -- hex spatial augmentation


def test_hex_rotate():
    hexrotate = HexRotate(extent=1, n_rot=0)
    sequence = torch.tensor(np.arange(7)[None, None], dtype=torch.float32)
    assert_equals = {
        0: [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        1: [1.0, 4.0, 0.0, 3.0, 6.0, 2.0, 5.0],
        2: [4.0, 6.0, 1.0, 3.0, 5.0, 0.0, 2.0],
        3: [6.0, 5.0, 4.0, 3.0, 2.0, 1.0, 0.0],
        4: [5.0, 2.0, 6.0, 3.0, 0.0, 4.0, 1.0],
        5: [2.0, 0.0, 5.0, 3.0, 1.0, 6.0, 4.0],
    }

    for n_rot, expected in assert_equals.items():
        assert np.allclose(hexrotate(sequence, n_rot).cpu().numpy().flatten(), expected)

    def _angle(flow):
        return (np.degrees(np.arctan2(flow[1], flow[0])).mean() + 360) % 360

    flow = torch.tensor(
        np.repeat([[np.cos(np.radians(0))], [np.sin(np.radians(0))]], 7, axis=1)[None],
        dtype=torch.float32,
    )
    assert_equals = {0: 0, 1: 60, 2: 120, 3: 180, 4: 240, 5: 300}
    for n_rot, expected in assert_equals.items():
        hexrotate.n_rot = n_rot
        assert np.allclose(_angle(hexrotate(flow).cpu().numpy().squeeze()), expected)

    hexrotate.augment = False
    assert np.allclose(hexrotate(flow).tolist(), flow.tolist())


def test_hex_flip():
    hexflip = HexFlip(extent=1, axis=0)
    sequence = torch.tensor(np.arange(7)[None, None], dtype=torch.float32)
    assert_equals = {
        0: [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        1: [0.0, 2.0, 1.0, 3.0, 5.0, 4.0, 6.0],
        2: [5.0, 6.0, 2.0, 3.0, 4.0, 0.0, 1.0],
        3: [4.0, 1.0, 6.0, 3.0, 0.0, 5.0, 2.0],
    }

    for axis, expected in assert_equals.items():
        assert np.allclose(hexflip(sequence, axis).cpu().numpy().flatten(), expected)

    def _angle(flow):
        return (np.degrees(np.arctan2(flow[1], flow[0])).mean() + 360) % 360

    flow = torch.tensor(
        np.repeat([[np.cos(np.radians(0))], [np.sin(np.radians(0))]], 7, axis=1)[None],
        dtype=torch.float32,
    )
    assert_equals = {0: 0, 1: 180, 2: 300, 3: 60}
    for axis, expected in assert_equals.items():
        assert np.allclose(_angle(hexflip(flow, axis).cpu().numpy().squeeze()), expected)
    hexflip.augment = False
    assert np.allclose(hexflip(flow).tolist(), flow.tolist())


def test_jitter():
    jitter = ContrastBrightness(0.5, 0.5)
    sequence = torch.tensor(
        np.array([0, 1, 1, 1, 1, 0, 0])[None, None], dtype=torch.float32
    )
    assert np.allclose(
        jitter(sequence).cpu().numpy().flatten(),
        [0.5, 1.0, 1.0, 1.0, 1.0, 0.5, 0.5],
    )
    sequence = torch.tensor(
        np.array([-1, 1, 1, 1, 1, 0, 0])[None, None], dtype=torch.float32
    )
    assert np.allclose(
        jitter(sequence).cpu().numpy().flatten(),
        [0.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.5],
    )
    jitter.augment = False
    assert np.allclose(jitter(sequence).tolist(), sequence.tolist())


def test_noise():
    noise = PixelNoise(0.5)
    sequence = torch.tensor(
        np.random.random_sample(size=[1, 1, 721]), dtype=torch.float32
    )
    transformed = noise(sequence)
    snr_light = torch.mean(transformed[sequence > 0.5]) ** 2 / torch.var(
        transformed[sequence > 0.5]
    )
    snr_dark = torch.mean(transformed[sequence <= 0.5]) ** 2 / torch.var(
        transformed[sequence <= 0.5]
    )
    assert snr_light > snr_dark
    noise.augment = False
    assert np.allclose(noise(sequence).tolist(), sequence.tolist())


# -- temporal augmentation


def test_random_crop():
    def _is_monotonous(sequence):
        return np.all(np.diff(sequence) >= 0)

    random_crop = CropFrames(10, start=0, all_frames=False, random=True)
    sequence = torch.arange(1000)

    # fixed start frame at 0
    cropped1 = random_crop(sequence).cpu().numpy().flatten()
    cropped2 = random_crop(sequence).cpu().numpy().flatten()
    assert len(cropped1) == 10 == len(cropped2)
    assert _is_monotonous(cropped1) and _is_monotonous(cropped2)
    assert cropped1.min() in sequence and cropped1.max() in sequence
    assert cropped2.min() in sequence and cropped2.max() in sequence
    assert np.allclose(cropped1, cropped2)

    # random start frame
    random_crop.set_or_sample(start=None, total_sequence_length=len(sequence))
    cropped3 = random_crop(sequence).cpu().numpy().flatten()
    assert not np.allclose(cropped1, cropped3)

    # provided start frame
    random_crop.set_or_sample(start=42, total_sequence_length=len(sequence))
    cropped4 = random_crop(sequence).cpu().numpy().flatten()
    assert not np.allclose(cropped1, cropped4)

    # turn augmentation on and off
    random_crop.augment = False
    assert np.allclose(random_crop(sequence).tolist(), sequence.tolist())
    random_crop.augment = True
    random_crop.all_frames = True
    assert np.allclose(random_crop(sequence).tolist(), sequence.tolist())
    random_crop.random = False
    random_crop.all_frames = False
    assert np.allclose(random_crop(sequence).tolist(), list(range(10)))
    random_crop.random = True
    random_crop.set_or_sample(start=None, total_sequence_length=len(sequence))
    random_crop.all_frames = False
    assert not np.allclose(random_crop(sequence).tolist(), list(range(10)))


def test_resample():
    resample = Interpolate(24, 50, mode="nearest-exact")
    sequence = torch.arange(10)[None, None]
    resampled = resample(sequence, dim=2).flatten().tolist()
    assert np.allclose(
        resampled,
        [0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9],
    )
    assert np.allclose(
        resampled,
        sequence.flatten()[
            resample.piecewise_constant_indices(len(sequence.flatten()))
        ].tolist(),
    )
    assert len(resampled) == math.ceil(10 * 50 / 24)
    resample.augment = False
    assert np.allclose(resample(sequence).flatten().tolist(), sequence.tolist())


def test_interpolate():
    interpolate = Interpolate(24, 50, mode="linear")
    flow = torch.tensor(
        [
            [np.cos(np.radians(0)), np.sin(np.radians(0))],
            [np.cos(np.radians(90)), np.sin(np.radians(90))],
            [np.cos(np.radians(180)), np.sin(np.radians(180))],
        ],
        dtype=torch.float32,
    )[:, :, None]
    interpolated = interpolate(flow)
    target = [
        [1, 0],
        [2 / 3, 1 / 3],
        [1 / 3, 2 / 3],
        [0, 1],
        [-1 / 3, 2 / 3],
        [-2 / 3, 1 / 3],
        [-1, 0],
    ]
    assert len(interpolated) == 7
    assert np.allclose(interpolated.cpu().numpy().squeeze(), target)
    interpolate.augment = False
    assert np.allclose(interpolate(flow).tolist(), flow.tolist())
